import numpy as np

import openmdao.api as om


class TrueAirspeedComp(om.ExplicitComponent):
    """ Compute Mach number based on true airspeed and the local speed of sound. """

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']
        self.add_input('sos', val=np.zeros(nn), desc='atmospheric speed of sound', units='m/s')
        self.add_input('mach', val=np.zeros(nn), desc='Mach number', units=None)
        self.add_output('TAS', val=np.zeros(nn), desc='true airspeed', units='m/s')

        # Setup partials
        ar = np.arange(self.options['num_nodes'])

        self.declare_partials(of='TAS', wrt='mach', rows=ar, cols=ar)
        self.declare_partials(of='TAS', wrt='sos', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        outputs['TAS'] = inputs['mach'] * inputs['sos']

    def compute_partials(self, inputs, partials):
        partials['TAS', 'mach'] = inputs['sos']
        partials['TAS', 'sos'] = inputs['mach']



from collections.abc import Sequence

import numpy as np

import openmdao.api as om
from openmdao.utils.om_warnings import warn_deprecation

from .explicit_shooting_continuity_comp import ExplicitShootingContinuityComp
from ..transcription_base import TranscriptionBase
from ..grid_data import BirkhoffGrid, GaussLobattoGrid, RadauGrid, UniformGrid, ChebyshevGaussLobattoGrid
from .ode_integration_comp import ODEIntegrationComp
from ...utils.misc import get_rate_units, CoerceDesvar
from ...utils.indexing import get_src_indices_by_row
from ...utils.introspection import get_promoted_vars, get_source_metadata, get_targets, _get_targets_metadata
from ...utils.constants import INF_BOUND

from ...utils.ode_utils import _make_ode_system
from ..common import TimeComp, TimeseriesOutputComp, ControlInterpComp, ParameterComp


class ExplicitShooting(TranscriptionBase):
    r"""
    The Transcription class for single explicit shooting.

    This transcription uses an external explicit integrator to propagate the states, and optionally their
    sensitivities wrt the 'inputs' to the integration.

    If we view integration as a function

    .. math::
        \bar{x}_{f} = \mathcal{I}(\bar{x}_0, t_0, t_d, \bar{\theta}) \\
        \bar{x}_{f} = \bar{x}_{0} + \int_{t_0}^{t_0+t_d} \left( f_{ode}(\bar{x}, t, \bar{\theta}) \right) dt

    then the inputs are the initial states ($\bar{x}$), the initial time and duration ($t_0$ and $t_d$), and some set
    of parameters that impact the ODE ($\theta$). For Dymos, $\theta$ may include the phase parameters, or the node values
    that govern the shape of the controls.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of arguments.
    """  # nopep8: E501, W605
    def __init__(self, **kwargs):
        super(ExplicitShooting, self).__init__(**kwargs)
        self._rhs_source = 'ode'

    def initialize(self):
        """
        Declare transcription options.
        """
        self.options.declare('method', types=str, default='DOP853',
                             desc='The integration method used.')
        self.options.declare('atol', types=float, default=1.0E-6)
        self.options.declare('rtol', types=float, default=1.0E-9)
        self.options.declare('first_step', types=float, allow_none=True, default=None)
        self.options.declare('max_step', types=float, default=np.inf)
        self.options.declare('propagate_derivs', types=bool, default=True,
                             desc='If True, propagate the state and derivatives of the state and time with respect to '
                                  'the integration parameters. If False, only propagate the primal states. If only '
                                  'using this transcription to propagate an ODE and derivatives are nof needed, '
                                  'setting this option to False should result in faster execution.')
        self.options.declare('subprob_reports', default=False,
                             desc='Controls the reports made when running the subproblems for ExplicitShooting')
        self.options.declare('grid', types=(GaussLobattoGrid, ChebyshevGaussLobattoGrid,
                                            RadauGrid, BirkhoffGrid, str), allow_none=True, default=None,
                             desc='The grid distribution used to layout the control inputs and provide the default '
                                  'output nodes.')
        self.options.declare('output_grid', types=(GaussLobattoGrid, ChebyshevGaussLobattoGrid, RadauGrid,
                                                   UniformGrid, BirkhoffGrid), allow_none=True,
                             default=None,
                             desc='The grid distribution determining the location of the output nodes. The default '
                                  'value of None will result in the use of the grid for outputs. This is useful '
                                  'for the implementation of path constraints but can result in highly nonlinear '
                                  'dynamics being smoothed over in the outputs. When used for validation through '
                                  'simulation, it is generally wise to choose an output grid that is more dense '
                                  'than the input grid to capture this nonlinearity.')
        self.options.declare('num_steps_per_segment', types=int, allow_none=True,
                             default=None, desc='Number of integration steps in each segment',
                             deprecation='Option `num_steps_per_segment is deprecated. ExplicitShooting now uses '
                                         'adaptive-step methods.')
        self.options.declare('control_interp', values=['vandermonde', 'barycentric', 'cubic'], default='vandermonde',
                             desc='Control interpolation algorithm, one of either "vandermonde", "barycentric",'
                             ' or "cubic". In general, Vandermonde is faster but Barycentric is necessary for the'
                             ' Birkhoff transcription where the number of nodes per segment can exceed 20 to 30.'
                             ' Cubic uses the scipy interpolate cubic spline method and is typically the fastest'
                             ' of the three')

        # Deprecated options previously inherited from transcription base.
        self.options.declare('num_segments', types=int, desc='Number of segments')
        self.options.declare('segment_ends', default=None, types=(Sequence, np.ndarray),
                             allow_none=True, desc='Locations of segment ends or None for equally '
                             'spaced segments')
        self.options.declare('order', default=3, types=(int, Sequence, np.ndarray),
                             desc='Order of the state transcription. The order of the control '
                                  'transcription is `order - 1`.')
        self.options.declare('compressed', default=True, types=bool,
                             desc='Use compressed transcription, meaning state and control values'
                                  'at segment boundaries are not duplicated on input.  This '
                                  'implicitly enforces value continuity between segments but in '
                                  'some cases may make the problem more difficult to solve.')

    def init_grid(self):
        """
        Setup the GridData object for the Transcription.
        """
        if self.options['grid'] in ('gauss-lobatto', None):
            self.options['grid'] = GaussLobattoGrid(num_segments=self.options['num_segments'],
                                                    nodes_per_seg=self.options['order'],
                                                    segment_ends=self.options['segment_ends'],
                                                    compressed=self.options['compressed'])
        elif self.options['grid'] == 'radau-ps':
            self.options['grid'] = RadauGrid(num_segments=self.options['num_segments'],
                                             nodes_per_seg=self.options['order'] + 1,
                                             segment_ends=self.options['segment_ends'],
                                             compressed=self.options['compressed'])

        dep_methods = {'rk4', '3/8', 'euler', 'ralston', 'rkf', 'rkck', 'dopri'}
        if self.options['method'] in dep_methods:
            warn_deprecation(f'Integration method {self.options["method"]} is no longer a valid option. Please use one '
                             f'of \'DOP853\', \'RK45\', \'RK23\' instead. Falling back to the default \'DOP853\'.')
            self.options['method'] = 'DOP853'
        # End deprecation-handling

        self.grid_data = self.options['grid']

        if self.options['output_grid']:
            self._output_grid_data = self.options['output_grid']
        else:
            self._output_grid_data = self.grid_data

    def setup_time(self, phase):
        """
        Setup the time component.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        time_options = phase.time_options
        t_name = time_options['name']
        t_phase_name = f'{t_name}_phase'
        time_units = time_options['units']

        # Warn about invalid options
        phase.check_time_options()

        for ts_name, ts_options in phase._timeseries.items():
            if t_name not in ts_options['outputs']:
                phase.add_timeseries_output(t_name, timeseries=ts_name)
            if t_phase_name not in ts_options['outputs'] and phase.timeseries_options['include_t_phase']:
                phase.add_timeseries_output(t_phase_name, timeseries=ts_name)

        # Case 1:  Compute times at 'all' node set.
        num_nodes = self._output_grid_data.num_nodes
        node_ptau = self._output_grid_data.node_ptau
        node_dptau_dstau = self._output_grid_data.node_dptau_dstau

        time_comp = TimeComp(num_nodes=num_nodes, node_ptau=node_ptau,
                             node_dptau_dstau=node_dptau_dstau, units=time_units)

        phase.add_subsystem('param_comp', subsys=ParameterComp(time_options=time_options),
                            promotes_inputs=['*'], promotes_outputs=['*'])

        phase.add_subsystem('time', time_comp, promotes_outputs=['*'])

        phase.connect('t_initial_val', ['time.t_initial', 'integrator.t_initial'])
        phase.connect('t_duration_val', ['time.t_duration', 'integrator.t_duration'])

    def configure_time(self, phase):
        """
        Configure the inputs/outputs on the time component.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        time_options = phase.time_options
        t_name = time_options['name']
        tphase_name = f'{t_name}_phase'

        integ = phase._get_subsystem('integrator')
        integ._configure_time()

        time_comp = phase._get_subsystem('time')
        time_comp.configure_io()

        ode = phase._get_subsystem('ode')

        if not time_options['fix_initial']:
            lb, ub = time_options['initial_bounds']
            lb = -INF_BOUND if lb is None else lb
            ub = INF_BOUND if ub is None else ub

            phase.add_design_var('t_initial',
                                 lower=lb,
                                 upper=ub,
                                 scaler=time_options['initial_scaler'],
                                 adder=time_options['initial_adder'],
                                 ref0=time_options['initial_ref0'],
                                 ref=time_options['initial_ref'])

        if not (time_options['input_duration'] or time_options['fix_duration']):
            lb, ub = time_options['duration_bounds']
            lb = -INF_BOUND if lb is None else lb
            ub = INF_BOUND if ub is None else ub

            phase.add_design_var('t_duration',
                                 lower=lb,
                                 upper=ub,
                                 scaler=time_options['duration_scaler'],
                                 adder=time_options['duration_adder'],
                                 ref0=time_options['duration_ref0'],
                                 ref=time_options['duration_ref'])

        # The tuples here are (name, user_specified_targets, dynamic)
        for name, targets, dynamic in [(t_name, time_options['targets'], True),
                                       (tphase_name, time_options['time_phase_targets'], True)]:
            if targets:
                src_idxs = self._output_grid_data.subset_node_indices['all'] if dynamic else None
                phase.connect(f'integrator.{name}', [f'ode.{t}' for t in targets], src_indices=src_idxs,
                              flat_src_indices=True if dynamic else None)

        for name, tgts in [('t_initial', time_options['t_initial_targets']),
                           ('t_duration', time_options['t_duration_targets'])]:

            targets = _get_targets_metadata(ode, name, user_targets=tgts)
            for t, meta in targets.items():
                tgt_shape = meta['shape']

                if tgt_shape == (1,):
                    src_idxs = None
                    flat_src_idxs = None
                else:
                    src_idxs = np.zeros(self._output_grid_data.subset_num_nodes['all'])
                    flat_src_idxs = True

                phase.connect(f'{name}_val', f'ode.{name}',
                              src_indices=src_idxs,
                              flat_src_indices=flat_src_idxs)

    def setup_states(self, phase):
        """
        Setup the states for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def configure_states(self, phase):
        """
        Configure state connections post-introspection.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        integ = phase._get_subsystem('integrator')
        integ._configure_states()

        state_prefix = 'states:' if phase.timeseries_options['use_prefix'] else ''

        for name, options in phase.state_options.items():
            phase.promotes('integrator', inputs=[(f'states:{name}', f'initial_states:{name}')])
            for ts_name, ts_options in phase._timeseries.items():
                if f'{state_prefix}{name}' not in ts_options['outputs']:
                    phase.add_timeseries_output(name, output_name=f'{state_prefix}{name}',
                                                timeseries=ts_name)

        # Add the appropriate design variables
        for state_name, options in phase.state_options.items():
            if options['fix_final']:
                raise ValueError('fix_final is not a valid option for states when using the '
                                 'ExplicitShooting transcription.')
            if options['opt'] and not options['fix_initial']:
                phase.add_design_var(name=f'initial_states:{state_name}',
                                     lower=options['lower'],
                                     upper=options['upper'],
                                     scaler=options['scaler'],
                                     adder=options['adder'],
                                     ref0=options['ref0'],
                                     ref=options['ref'])

    def setup_ode(self, phase):
        """
        Setup the ode for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        integ = ODEIntegrationComp(ode_class=phase.options['ode_class'],
                                   time_options=phase.time_options,
                                   state_options=phase.state_options,
                                   parameter_options=phase.parameter_options,
                                   control_options=phase.control_options,
                                   method=self.options['method'],
                                   atol=self.options['atol'],
                                   rtol=self.options['rtol'],
                                   first_step=self.options['first_step'],
                                   max_step=self.options['max_step'],
                                   propagate_derivs=self.options['propagate_derivs'],
                                   input_grid_data=self.grid_data,
                                   output_grid_data=self._output_grid_data,
                                   ode_init_kwargs=phase.options['ode_init_kwargs'],
                                   standalone_mode=False,
                                   reports=self.options['subprob_reports'],
                                   control_interp=self.options['control_interp'],
                                   calc_exprs=phase._calc_exprs)
        phase.add_subsystem('integrator', integ)

        ode = _make_ode_system(ode_class=phase.options['ode_class'],
                               num_nodes=self._output_grid_data.num_nodes,
                               calc_exprs=phase._calc_exprs,
                               ode_init_kwargs=phase.options['ode_init_kwargs'],
                               parameter_options=phase.parameter_options)

        phase.add_subsystem('ode', ode)

    def configure_ode(self, phase):
        """
        Create connections to the introspected states.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        integ = phase._get_subsystem('integrator')
        integ._configure_storage()

        ode = phase._get_subsystem('ode')

        ode_inputs = get_promoted_vars(ode, 'input')

        for name, options in phase.state_options.items():

            targets = get_targets(ode_inputs, name=name, user_targets=options['targets'])
            if targets:
                phase.connect(f'integrator.states_out:{name}',
                              [f'ode.{tgt}' for tgt in targets])

    def setup_controls(self, phase):
        """
        Setup the control group.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        phase._check_control_options()

        if phase.control_options:
            control_comp = ControlInterpComp(control_options=phase.control_options,
                                             time_units=phase.time_options['units'],
                                             grid_data=self.options['grid'],
                                             output_grid_data=self._output_grid_data)

            phase.add_subsystem('control_comp',
                                subsys=control_comp,
                                promotes=[('t_duration', 't_duration_val'), 'dt_dstau',
                                          '*controls:*', '*control_values:*', '*control_rates:*'])

            control_prefix = 'controls:' if phase.timeseries_options['use_prefix'] else ''
            control_rate_prefix = 'control_rates:' if phase.timeseries_options['use_prefix'] else ''

            for name, options in phase.control_options.items():
                for ts_name, ts_options in phase._timeseries.items():

                    if f'{control_prefix}{name}' not in ts_options['outputs']:
                        phase.add_timeseries_output(name, output_name=f'{control_prefix}{name}',
                                                    timeseries=ts_name)
                    if f'{control_rate_prefix}{name}_rate' not in ts_options['outputs'] \
                            and phase.timeseries_options['include_control_rates']:
                        phase.add_timeseries_output(f'{name}_rate', output_name=f'{control_rate_prefix}{name}_rate',
                                                    timeseries=ts_name)
                    if f'{control_rate_prefix}{name}_rate2' not in ts_options['outputs'] \
                            and phase.timeseries_options['include_control_rates']:
                        phase.add_timeseries_output(f'{name}_rate2', output_name=f'{control_rate_prefix}{name}_rate2',
                                                    timeseries=ts_name)

    def configure_controls(self, phase):
        """
        Configure the inputs/outputs for the controls.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        if not phase.control_options:
            return

        integrator_comp = phase._get_subsystem('integrator')
        integrator_comp._configure_controls()
        control_comp = phase._get_subsystem('control_comp')
        control_comp.configure_io()

        ode = phase._get_subsystem('ode')
        ode_inputs = get_promoted_vars(ode, 'input')

        # Add the appropriate design parameters
        for control_name, options in phase.control_options.items():
            if options['control_type'] == 'polynomial':
                ncin = options['order'] + 1
            else:
                ncin = self.options['grid'].subset_num_nodes['control_input']

            phase.promotes('integrator', inputs=[f'controls:{control_name}'])

            if options['opt']:
                coerce_desvar_option = CoerceDesvar(num_input_nodes=ncin, options=options)

                phase.add_design_var(name=f'controls:{control_name}',
                                     lower=coerce_desvar_option('lower'),
                                     upper=coerce_desvar_option('upper'),
                                     scaler=coerce_desvar_option('scaler'),
                                     adder=coerce_desvar_option('adder'),
                                     ref0=coerce_desvar_option('ref0'),
                                     ref=coerce_desvar_option('ref'),
                                     indices=om.slicer[coerce_desvar_option.desvar_indices, ...])

            # Control targets are detected automatically
            targets = get_targets(ode_inputs, control_name, options['targets'])

            if targets:
                phase.connect(f'control_values:{control_name}',
                              [f'ode.{t}' for t in targets])

            # Rate targets
            rate_targets = get_targets(ode_inputs, control_name, options['rate_targets'])

            if rate_targets:
                phase.connect(f'control_rates:{control_name}_rate',
                              [f'ode.{t}' for t in rate_targets])

            # Second time derivative targets must be specified explicitly
            rate2_targets = get_targets(ode_inputs, control_name, options['rate2_targets'])

            if rate2_targets:
                phase.connect(f'control_rates:{control_name}_rate2',
                              [f'ode.{t}' for t in targets])

    def configure_parameters(self, phase):
        """
        Configure parameter promotion.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super().configure_parameters(phase)

        integrator_comp = phase._get_subsystem('integrator')
        integrator_comp._configure_parameters()

    def setup_defects(self, phase):
        """
        Create the continuity_comp to house the defects.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        state_cont, control_cont, rate_cont = self._requires_continuity_constraints(phase)

        if state_cont or control_cont or rate_cont:
            phase.add_subsystem('continuity_comp',
                                ExplicitShootingContinuityComp(grid_data=self.options['grid'],
                                                               state_options=phase.state_options,
                                                               control_options=phase.control_options,
                                                               time_units=phase.time_options['units']))

            if rate_cont:
                phase.connect('t_duration_val', 'continuity_comp.t_duration')

    def configure_defects(self, phase):
        """
        Not used in ExplicitShooting.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        ogd = self._output_grid_data
        any_state_cnty, any_control_cnty, any_rate_cnty = self._requires_continuity_constraints(phase)
        src_idxs = om.slicer[ogd.subset_node_indices['segment_ends'], ...]

        controls_to_enforce = set()
        control_rates_to_enforce = set()
        control_rates2_to_enforce = set()

        for control_name, options in phase.control_options.items():

            if options['continuity'] and any_control_cnty:
                controls_to_enforce.add(control_name)
                phase.connect(f'control_values:{control_name}',
                              f'continuity_comp.controls:{control_name}',
                              src_indices=src_idxs)
            if options['rate_continuity'] and any_rate_cnty:
                control_rates_to_enforce.add(control_name)
                phase.connect(f'control_rates:{control_name}_rate',
                              f'continuity_comp.control_rates:{control_name}_rate',
                              src_indices=src_idxs)
            if options['rate2_continuity'] and any_rate_cnty:
                control_rates2_to_enforce.add(control_name)
                phase.connect(f'control_rates:{control_name}_rate2',
                              f'continuity_comp.control_rates:{control_name}_rate2',
                              src_indices=src_idxs)

        if any((controls_to_enforce, control_rates_to_enforce, control_rates2_to_enforce)):
            phase.continuity_comp.configure_io(controls_to_enforce=controls_to_enforce,
                                               control_rates_to_enforce=control_rates_to_enforce,
                                               control_rates2_to_enforce=control_rates2_to_enforce)

    def setup_timeseries_outputs(self, phase):
        """
        Setup the timeseries for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        for name, options in phase._timeseries.items():
            timeseries_comp = TimeseriesOutputComp(input_grid_data=self._output_grid_data,
                                                   output_grid_data=self._output_grid_data,
                                                   output_subset=options['subset'],
                                                   time_units=phase.time_options['units'])
            phase.add_subsystem(name, subsys=timeseries_comp)

            phase.connect('dt_dstau', f'{name}.dt_dstau', flat_src_indices=True)

    def configure_timeseries_outputs(self, phase):
        """
        Create connections from time series to all post-introspection sources.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super().configure_timeseries_outputs(phase)

    def setup_solvers(self, phase):
        """
        Setup the solvers for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def configure_solvers(self, phase, requires_solvers=None):
        """
        Setup the solvers for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        requires_solvers : None
            Required to extend TranscriptionBase.configure_solvers but not used.
        """
        super().configure_solvers(phase, requires_solvers=None)

    def get_parameter_connections(self, name, phase):
        """
        Returns info about a parameter's target connections in the phase.

        Parameters
        ----------
        name : str
            Parameter name.
        phase : dymos.Phase
            The phase object to which this transcription instance applies.

        Returns
        -------
        list of (paths, indices)
            A list containing a tuple of target paths and corresponding src_indices to which the
            given design variable is to be connected.
        """
        connection_info = []

        if name in phase.parameter_options:
            options = phase.parameter_options[name]
            for tgt in options['targets']:
                if tgt in options['static_targets']:
                    src_idxs_raw = np.zeros(1, dtype=int)
                    src_idxs = get_src_indices_by_row(src_idxs_raw, options['shape'])
                    src_idxs = np.squeeze(src_idxs, axis=0)
                else:
                    src_idxs_raw = np.zeros(self._output_grid_data.subset_num_nodes['all'], dtype=int)
                    src_idxs = get_src_indices_by_row(src_idxs_raw, options['shape'])
                    if options['shape'] == (1,):
                        src_idxs = src_idxs.ravel()

                connection_info.append((f'ode.{tgt}', (src_idxs,)))
            connection_info.append(([f'integrator.parameters:{name}'], None))

        return connection_info

    def _get_response_src(self, var, loc, phase, ode_outputs=None):
        """
        Return the path to the variable that will be used as a response..

        Parameters
        ----------
        var : str
            Name of the variable to be used as the response.
        loc : str
            The location of the response in the phase ['initial', 'final'].
        phase : dymos.Phase
            Phase object containing in which the objective resides.
        ode_outputs : dict or None
            A dictionary of ODE outputs as returned by get_promoted_vars.

        Returns
        -------
        obj_path : str
            Path to the source.
        shape : tuple
            Source shape.
        units : str
            Source units.
        linear : bool
            True if the objective quantity is linear.
        """
        time_units = phase.time_options['units']
        var_type = phase.classify_var(var)

        if var_type == 't':
            shape = (1,)
            units = time_units
            linear = True
            obj_path = 't'
        elif var_type == 't_phase':
            shape = (1,)
            units = time_units
            linear = True
            obj_path = 't_phase'
        elif var_type == 'state':
            shape = phase.state_options[var]['shape']
            units = phase.state_options[var]['units']
            linear = loc == 'initial'
            obj_path = f'integrator.states_out:{var}'
        elif var_type == 'control':
            shape = phase.control_options[var]['shape']
            units = phase.control_options[var]['units']
            linear = False
            obj_path = f'control_values:{var}'
        elif var_type == 'parameter':
            shape = phase.parameter_options[var]['shape']
            units = phase.parameter_options[var]['units']
            linear = True
            obj_path = f'parameter_vals:{var}'
        elif var_type in ('control_rate', 'control_rate2'):
            control_var = var[:-5] if var_type == 'control_rate' else var[:-6]
            shape = phase.control_options[control_var]['shape']
            control_units = phase.control_options[control_var]['units']
            d = 2 if var_type == 'control_rate2' else 1
            control_rate_units = get_rate_units(control_units, time_units, deriv=d)
            units = control_rate_units
            linear = False
            obj_path = f'control_rates:{var}'
        else:
            # Failed to find variable, assume it is in the ODE. This requires introspection.
            obj_path = f'{self._rhs_source}.{var}'
            if ode_outputs is None:
                ode = self._get_ode(phase)
            else:
                ode = ode_outputs
            meta = get_source_metadata(ode, var, user_units=None, user_shape=None)
            shape = meta['shape']
            units = meta['units']
            linear = False

        return obj_path, shape, units, linear

    def _requires_continuity_constraints(self, phase):
        """
        Tests whether state and/or control and/or control rate continuity are required.

        Parameters
        ----------
        phase : dymos.Phase
            The phase to which this transcription applies.

        Returns
        -------
        state_continuity : bool
            True if any state continuity is required to be enforced.
        control_continuity : bool
            True if any control value continuity is required to be enforced.
        control_rate_continuity : bool
            True if any control rate continuity is required to be enforced.
        """
        num_seg = self.options['grid'].num_segments
        compressed = self.options['grid'].compressed
        transcription = self.options['grid'].transcription

        state_continuity = False
        any_control_continuity = any([opts['continuity'] for opts in phase.control_options.values()])
        any_control_continuity = any_control_continuity and num_seg > 1 and not (compressed or transcription == 'radau-ps')
        any_rate_continuity = any([opts['rate_continuity'] or opts['rate2_continuity']
                                   for opts in phase.control_options.values()])
        any_rate_continuity = any_rate_continuity and num_seg > 1

        return state_continuity, any_control_continuity, any_rate_continuity

    def _get_num_timeseries_nodes(self):
        """
        Returns the number of nodes in the default timeseries for this transcription.

        Returns
        -------
        int
            The number of nodes in the default timeseries for this transcription.
        """
        return self._output_grid_data.subset_num_nodes['all']

    def _get_timeseries_var_source(self, var, output_name, phase):
        """
        Return the source path and indices for a given variable to be connected to a timeseries.

        Parameters
        ----------
        var : str
            Name of the timeseries variable whose source is desired.
        output_name : str
            Name of the timeseries output whose source is desired.
        phase : dymos.Phase
            Phase object containing the variable, either as state, time, control, etc., or as an ODE output.

        Returns
        -------
        meta : dict
            Metadata pertaining to the variable at the given path. This dict contains 'src' (the path to the
            timeseries source), 'src_idxs' (an array of the
            source indices), 'units' (the units of the source variable), and 'shape' (the shape of the variable at
            a given node).
        """
        var_type = phase.classify_var(var)
        time_units = phase.time_options['units']
        time_name = phase.time_options['name']

        transcription = phase.options['transcription']
        ode = transcription._get_ode(phase)
        ode_outputs = get_promoted_vars(ode, 'output')

        # The default for node_idxs, applies to everything except states and parameters.
        node_idxs = None
        meta = {}

        # Determine the path to the variable
        if var_type == 't':
            path = f'integrator.{time_name}'
            src_units = time_units
            src_shape = (1,)
        elif var_type == 't_phase':
            path = f'integrator.{time_name}_phase'
            src_units = time_units
            src_shape = (1,)
        elif var_type == 'state':
            path = f'integrator.states_out:{var}'
            src_units = phase.state_options[var]['units']
            src_shape = phase.state_options[var]['shape']
        elif var_type in ['control']:
            path = f'control_values:{var}'
            src_units = phase.control_options[var]['units']
            src_shape = phase.control_options[var]['shape']
        elif var_type == 'control_rate':
            control_name = var[:-5]
            path = f'control_rates:{control_name}_rate'
            src_units = get_rate_units(phase.control_options[control_name]['units'], time_units, deriv=1)
            src_shape = phase.control_options[control_name]['shape']
        elif var_type == 'control_rate2':
            control_name = var[:-6]
            path = f'control_rates:{control_name}_rate2'
            src_units = get_rate_units(phase.control_options[control_name]['units'], time_units, deriv=2)
            src_shape = phase.control_options[control_name]['shape']
        elif var_type == 'parameter':
            path = f'parameter_vals:{var}'
            node_idxs = np.zeros(self._output_grid_data.subset_num_nodes['all'], dtype=int)
            src_units = phase.parameter_options[var]['units']
            src_shape = phase.parameter_options[var]['shape']
        else:
            # Failed to find variable, assume it is in the ODE
            meta = get_source_metadata(ode_outputs, src=var)
            src_shape = meta['shape']
            src_units = meta['units']
            src_tags = meta['tags']
            path = f'ode.{var}'
            if 'dymos.static_output' in src_tags:
                raise RuntimeError(f'ODE output {var} is tagged with "dymos.static_output" and cannot be a '
                                   f'timeseries output.')

        src_idxs = None if node_idxs is None else om.slicer[node_idxs, ...]

        meta['src'] = path
        meta['src_idxs'] = src_idxs
        meta['units'] = src_units
        meta['shape'] = src_shape

        return meta

    def _phase_set_state_val(self, phase, name, vals, time_vals=None, interpolation_kind=None):
        """
        Method to interpolate the provided input and return the variables that need to be set
        along with their appropriate value.

        Parameters
        ----------
        phase : dymos.Phase
            The phase to which this transcription applies.
        name : str
            The name of the phase variable to be set.
        vals : ndarray or Sequence or float
            Array of control/state/parameter values.
        time_vals : ndarray or Sequence or None
            Array of integration variable values.
        interpolation_kind : str or None
            Specifies the kind of interpolation, as per the scipy.interpolate package.
            One of ('linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic'
            where 'zero', 'slinear', 'quadratic' and 'cubic' refer to a spline
            interpolation of zeroth, first, second or third order) or as an
            integer specifying the order of the spline interpolator to use.
            Default is 'linear'.

        Returns
        -------
        input_data : dict
            Dict containing the values that need to be set in the phase

        """
        if np.isscalar(vals):
            val = vals
        else:
            val = vals[0]
        input_data = {f'initial_states:{name}': val}

        return input_data



import openmdao.api as om

from .carODE import CarODE
from .tireODE import TireODE
from .normalForceODE import NormalForceODE
from .accelerationODE import AccelerationODE
from .tireConstraintODE import TireConstraintODE
from .timeODE import TimeODE
from .timeAdder import TimeAdder
from .curvature import Curvature


class CombinedODE(om.Group):
    def initialize(self):
        self.options.declare("num_nodes", types=int,
                             desc="Number of nodes to be evaluated in the RHS")

    def setup(self):
        nn = self.options["num_nodes"]

        self.add_subsystem(name="normal", subsys=NormalForceODE(num_nodes=nn),
                           promotes_inputs=["ax", "ay", "V"],
                           promotes_outputs=["N_fr", "N_fl", "N_rr", "N_rl"], )

        self.add_subsystem(name="tire", subsys=TireODE(num_nodes=nn),
                           promotes_inputs=["lambda", "omega", "V", "N_fr", "N_fl", "N_rr", "N_rl",
                                            "thrust", "delta", ],
                           promotes_outputs=["S_fr", "S_fl", "S_rr", "S_rl", "F_fr", "F_fl", "F_rr",
                                             "F_rl", ], )

        self.add_subsystem(name="curv", subsys=Curvature(num_nodes=nn))

        self.connect("curv.kappa", "car.kappa")

        self.add_subsystem(name="car", subsys=CarODE(num_nodes=nn),
                           promotes_inputs=["n", "alpha", "V", "lambda", "omega", "S_fr", "S_fl",
                                            "S_rr", "S_rl", "F_fr", "F_fl", "F_rr", "F_rl",
                                            "delta", ],
                           promotes_outputs=["sdot", "ndot", "alphadot", "omegadot", "Vdot",
                                             "lambdadot", "power", ], )

        self.add_subsystem(name="accel", subsys=AccelerationODE(num_nodes=nn),
                           promotes_inputs=["V", "lambda", "omega", "Vdot", "lambdadot", "ax",
                                            "ay"], promotes_outputs=["axdot", "aydot"], )

        self.add_subsystem(name="tireconstraint", subsys=TireConstraintODE(num_nodes=nn),
                           promotes_inputs=["S_fr", "S_fl", "S_rr", "S_rl", "F_fr", "F_fl", "F_rr",
                                            "F_rl", "N_fr", "N_fl", "N_rr", "N_rl", ],
                           promotes_outputs=["c_rr", "c_rl", "c_fr", "c_fl"], )

        self.add_subsystem(name="time", subsys=TimeODE(num_nodes=nn),
                           promotes_inputs=["ndot", "sdot", "omegadot", "lambdadot", "Vdot",
                                            "axdot", "aydot", "alphadot", ],
                           promotes_outputs=["dn_ds", "dV_ds", "domega_ds", "dlambda_ds",
                                             "dalpha_ds", "dax_ds", "day_ds", ], )

        self.add_subsystem(name="timeAdder", subsys=TimeAdder(num_nodes=nn),
                           promotes_inputs=["sdot"], promotes_outputs=["dt_ds"], )



"""
ODE for example that shows how to use multiple phases in Dymos to model failure of a battery cell
in a simple electrical system.
"""
import numpy as np
import openmdao.api as om

from dymos.examples.battery_multibranch.batteries import Battery
from dymos.examples.battery_multibranch.motors import Motors


class BatteryODE(om.Group):

    def initialize(self):
        self.options.declare('num_nodes', default=1)
        self.options.declare('num_battery', default=3)
        self.options.declare('num_motor', default=3)

    def setup(self):
        num_nodes = self.options['num_nodes']
        num_battery = self.options['num_battery']
        num_motor = self.options['num_motor']

        self.add_subsystem(name='pwr_balance',
                           subsys=om.BalanceComp(name='I_Li', val=1.0*np.ones(num_nodes),
                                                 rhs_name='pwr_out_batt',
                                                 lhs_name='P_pack',
                                                 units='A', eq_units='W', lower=0.0, upper=50.))

        self.add_subsystem('battery', Battery(num_nodes=num_nodes, n_parallel=num_battery),
                           promotes_inputs=['SOC'],
                           promotes_outputs=['dXdt:SOC'])

        self.add_subsystem('motors', Motors(num_nodes=num_nodes, n_parallel=num_motor))

        self.connect('battery.P_pack', 'pwr_balance.P_pack')
        self.connect('motors.power_in_motor', 'pwr_balance.pwr_out_batt')
        self.connect('pwr_balance.I_Li', 'battery.I_Li')
        self.connect('battery.I_pack', 'motors.current_in_motor')

        self.nonlinear_solver = om.NewtonSolver(solve_subsystems=False, maxiter=20)
        self.linear_solver = om.DirectSolver()



import openmdao.api as om
import dymos as dm
from dymos.examples.brachistochrone.brachistochrone_vector_states_ode \
    import BrachistochroneVectorStatesODE


def brachistochrone_min_time(transcription='gauss-lobatto', num_segments=8, transcription_order=3,
                             grid_type='lgl', compressed=True, optimizer='SLSQP',
                             dynamic_simul_derivs=True, force_alloc_complex=False,
                             solve_segments=False, run_driver=True, simulate=False,
                             make_plots=False):
    p = om.Problem(model=om.Group())

    if optimizer == 'SNOPT':
        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = optimizer
        p.driver.opt_settings['Major iterations limit'] = 100
        p.driver.opt_settings['Major feasibility tolerance'] = 1.0E-6
        p.driver.opt_settings['Major optimality tolerance'] = 1.0E-6
        p.driver.opt_settings['iSumm'] = 6
    else:
        p.driver = om.ScipyOptimizeDriver()

    if dynamic_simul_derivs:
        p.driver.declare_coloring()

    if transcription == 'gauss-lobatto':
        transcription = dm.GaussLobatto(num_segments=num_segments,
                                        order=transcription_order,
                                        compressed=compressed)
        fix_final = not solve_segments
    elif transcription == 'radau-ps':
        transcription = dm.Radau(num_segments=num_segments,
                                 order=transcription_order,
                                 compressed=compressed)
        fix_final = not solve_segments
    elif transcription == 'birkhoff':
        transcription = dm.Birkhoff(num_nodes=25,
                                    grid_type=grid_type)
        fix_final = not solve_segments

    traj = dm.Trajectory()
    phase = dm.Phase(ode_class=BrachistochroneVectorStatesODE,
                     transcription=transcription)
    traj.add_phase('phase0', phase)

    p.model.add_subsystem('traj0', traj)

    phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='s')

    # can't fix final position if you're solving the segments

    phase.add_state('pos', fix_initial=True, fix_final=fix_final,
                    solve_segments=solve_segments, ref=[1, 1])
    #
    phase.add_state('v', fix_initial=True, fix_final=False, solve_segments=solve_segments)
    #
    phase.add_control('theta',
                      continuity=True, rate_continuity=True,
                      units='deg', lower=0.01, upper=179.9)

    phase.add_parameter('g', opt=False, units='m/s**2', val=9.80665)

    if not fix_final:
        phase.add_boundary_constraint('pos', loc='final', equals=[10, 5])

    # Minimize time at the end of the phase
    phase.add_objective('time', loc='final', scaler=10)

    p.model.linear_solver = om.DirectSolver()
    p.setup(check=True, force_alloc_complex=force_alloc_complex)

    phase.set_time_val(initial=0.0, duration=1.8016)

    pos0 = [0, 10]
    posf = [10, 5]

    phase.set_state_val('pos', [pos0, posf])
    phase.set_state_val('v', [0, 9.9])
    phase.set_control_val('theta', [5, 100])
    phase.set_parameter_val('g', 9.80665)

    dm.run_problem(p,
                   run_driver=run_driver,
                   simulate=simulate,
                   make_plots=make_plots)

    return p


if __name__ == '__main__':
    p = brachistochrone_min_time(transcription='radau-ps', num_segments=5, run_driver=True,
                                 transcription_order=5, compressed=False, optimizer='SLSQP',
                                 solve_segments=False, force_alloc_complex=True,
                                 dynamic_simul_derivs=True)



import numpy as np
import openmdao.api as om
from ...grid_data import GridData
from ...._options import options as dymos_options


class ControlEndpointDefectComp(om.ExplicitComponent):
    r"""
    Compute/enforce the control endpoint defect when using the Radau Pseudospectral method.

    For each dynamic control, take the control values at all nodes.  Use a Radau interpolation
    matrix to compute the terminal value of the control, and compute the difference between this
    interpolated terminal value and the given terminal value.  Apply a constraint such that this
    value is zero.

    .. math:: {u}_{f} = \left[ L_{LGR} \right] vec{u}_{col}

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._no_check_partials = not dymos_options['include_check_partials']

    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare(
            'grid_data', types=GridData,
            desc='Container object for grid info')

        self.options.declare(
            'control_options', types=dict,
            desc='Dictionary of control names/options for the phase')

    def setup(self):
        """
        Add inputs/outputs for this component.
        """
        gd = self.options['grid_data']
        num_nodes = gd.subset_num_nodes['all']

        control_options = self.options['control_options']

        # We only need the last row of the Lagrange interpolation matrix since we
        # desire the interpolated control value at the endpoint of the phase.
        L, _ = gd.phase_lagrange_matrices('col', 'all')
        self._num_disc_end_segment = gd.subset_num_nodes_per_segment['col'][-1]
        self._L = np.atleast_2d(L[-1, -self._num_disc_end_segment:])

        self._input_str = {}
        self._output_str = {}

        for control_name, options in control_options.items():
            if not options['opt']:
                continue

            shape = options['shape']
            size = np.prod(shape)
            units = options['units']
            self._input_str[control_name] = f'controls:{control_name}'
            self._output_str[control_name] = f'control_endpoint_defects:{control_name}'

            self.add_input(name=self._input_str[control_name],
                           shape=(num_nodes,) + shape,
                           desc=f'Values of control {control_name} at all nodes',
                           units=units)

            self.add_output(
                name=self._output_str[control_name],
                shape=(1,) + shape,
                units=units,
                desc=f'Interpolated value of {control_name} at endpoint')

            self.add_constraint(name=self._output_str[control_name],
                                equals=0.0,
                                scaler=1.0,
                                linear=False)

            row_idxs = np.repeat(np.arange(size, dtype=int),
                                 gd.subset_num_nodes_per_segment['all'][-1])

            initial_col = sum([gd.subset_num_nodes_per_segment['all'][i] * size
                               for i in range(gd.num_segments - 1)])
            final_col = sum([gd.subset_num_nodes_per_segment['all'][i] * size
                             for i in range(gd.num_segments)])

            col_idxs = []
            for i in range(size):
                col_idxs.extend(np.arange(initial_col + i, final_col, size, dtype=int))

            dout_din = np.tile(np.concatenate((np.ravel(-self._L), [1.0])), size)

            self.declare_partials(of=self._output_str[control_name],
                                  wrt=self._input_str[control_name],
                                  rows=row_idxs, cols=col_idxs, val=dout_din)

    def compute(self, inputs, outputs):
        """
        Compute control endpoint defects.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        gd = self.options['grid_data']

        for name in self._input_str:
            u_col = inputs[self._input_str[name]][gd.subset_node_indices['col']]
            u_col_end_segment = np.atleast_2d(u_col[-self._num_disc_end_segment:])
            np.set_printoptions(linewidth=1024)
            outputs[self._output_str[name]] = inputs[self._input_str[name]][-1, :] - \
                np.tensordot(self._L, u_col_end_segment, axes=(1, 0))



import unittest

import openmdao.api as om
import dymos as dm


class ODENoTaggedStateUnits(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # Inputs
        self.add_input('v', shape=(nn,), desc='velocity', units='m/s')

        self.add_input('g', val=9.80665, desc='grav. acceleration', units='m/s/s', tags=['dymos.static_target'])

        self.add_input('theta', shape=(nn,), desc='angle of wire', units='rad')

        self.add_output('xdot', shape=(nn,), desc='velocity component in x', units='m/s',
                        tags=['dymos.state_rate_source:x'])

        self.add_output('ydot', shape=(nn,), desc='velocity component in y', units='m/s',
                        tags=['dymos.state_rate_source:y', 'dymos.state_units:m'])

        self.add_output('vdot', shape=(nn,), desc='acceleration magnitude', units='m/s**2',
                        tags=['dymos.state_rate_source:v', 'dymos.state_units:m/s'])

    def compute(self, inputs, outputs):
        pass

    def compute_partials(self, inputs, partials):
        pass


class TestTimeUnits(unittest.TestCase):

    def test_time_units_none(self):
        p = om.Problem(model=om.Group())

        transcription = dm.Radau(num_segments=10, order=3, compressed=True)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=ODENoTaggedStateUnits,
                         transcription=transcription)
        traj.add_phase('phase0', phase)

        p.model.add_subsystem('traj0', traj)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units=None)

        with self.assertRaises(RuntimeError) as e:
            p.setup(check=True, force_alloc_complex=False)

        expected = "Unable to infer the units of state variable `x` from\nthe rate units because the " \
                   "time units of the phase are set to None.\nChange the time units to something other " \
                   "than None, or explicitly\nset the state units using one of the following options:\n" \
                   "- Tag the state rate source `xdot` with `dymos.state_units:{units}`\n" \
                   "- Use the `set_state_options('x', units={units})` method on the phase."

        self.assertEqual(str(e.exception.__cause__), expected)



import fnmatch
from numbers import Number

import openmdao.api as om
import numpy as np
from openmdao.utils.units import simplify_unit
from openmdao.utils.general_utils import ensure_compatible
from dymos.utils.misc import _unspecified, _none_or_unspecified
from ..phase.options import StateOptionsDictionary, TimeseriesOutputOptionsDictionary
from .misc import get_rate_units


def classify_var(var, time_options, state_options, parameter_options, control_options,
                 timeseries_options=None):
    """
    Classifies a variable of the given name or path.

    This method searches for it as a time variable, state variable,
    control variable, or parameter.  If it is not found to be one
    of those variables, it is assumed to be the path to a variable
    relative to the top of the ODE system for the phase.

    Parameters
    ----------
    var : str
        The name of the variable to be classified.
    time_options : OptionsDictionary
        Time options for the phase.
    state_options : dict of {str: OptionsDictionary}
        For each state variable, a dictionary of its options, keyed by name.
    parameter_options : dict of {str: OptionsDictionary}
        For each parameter, a dictionary of its options, keyed by name.
    control_options : dict of {str: OptionsDictionary}
        For each control variable, a dictionary of its options, keyed by name.
    timeseries_options : {str: OptionsDictionary}
        For each timeseries, a dictionary of its options, keyed by name.

    Returns
    -------
    str
        The classification of the given variable, which is one of
        't', 't_phase', 'state', 'control', 'control_rate',
        'control_rate2', 'parameter', or 'ode'.
    """
    time_name = time_options['name']
    if var == time_name:
        return 't'
    elif var == f'{time_name}_phase':
        return 't_phase'
    elif var == 'time_phase':
        om.issue_warning('time_phase is deprecated. Please use `t_phase` to obtain the change in the integration '
                         'variable within the current phase.', category=om.OMDeprecationWarning)
        return 't_phase'
    elif var.startswith('initial_states:'):
        return 'state'
    elif var.startswith('final_states:'):
        return 'state'
    elif var in state_options:
        return 'state'
    elif var in control_options:
        return 'control'
    elif var in parameter_options:
        return 'parameter'
    elif var.endswith('_rate'):
        if var[:-5] in control_options:
            return 'control_rate'
    elif var.endswith('_rate2'):
        if var[:-6] in control_options:
            return 'control_rate2'
    return 'ode'


def get_promoted_vars(ode, iotypes, metadata_keys=None, get_remote=True):
    """
    Returns a dictionary mapping the promoted names of all inputs in a system to their associated metadata.

    Parameters
    ----------
    ode : openmdao.core.System
        The system from which the promoted inputs or outputs are being retrieved.
    iotypes : str or tuple
        One of 'input' or 'output', or a tuple of both.
    metadata_keys : Iterable or None
        Additional metadata requested for the variables. By default returns metadata available on 'allprocs'.  See
        openmdao.core.System.get_io_metadata for more information.
    get_remote : bool
        If True, include IO not local to this proc.

    Returns
    -------
    dict
        A dictionary mapping the promoted names of inputs in the system to their associated metadata.
    """
    _iotypes = (iotypes,) if isinstance(iotypes, str) else iotypes
    return {opts['prom_name']: opts for opts in ode.get_io_metadata(iotypes=_iotypes, get_remote=get_remote,
                                                                    metadata_keys=metadata_keys).values()}


def get_targets(ode, name, user_targets):
    """
    Return the targets of a variable in a given ODE system.

    If the targets of the variable is _unspecified, and the name is a top level input name
    in the ODE, then the values are automatically connected to that top-level input.
    If _unspecified and not a top-level input of the ODE, no connection is made.
    If targets is explicitly None, then no connection is made.
    Otherwise, if the user specified some other string or sequence of strings as targets, then
    those are returned.

    Parameters
    ----------
    ode : openmdao.core.System or dict
        The OpenMDAO system which serves as the ODE for dymos, or a dictionary of promoted ODE input names and
        their associated metadata. If a System, it should already have had its setup and configure methods called.
    name : str
        The name of the variable whose targets are desired.
    user_targets : str or None or Sequence or _unspecified
        Targets for the variable as given by the user.

    Returns
    -------
    list
        The target inputs of the variable in the ODE, as a list.

    Notes
    -----
    This method requires that the ODE has run its setup and configure methods.  Thus,
    this method should be called from configure of some parent Group, and the ODE should
    be a system within that Group.
    """
    if isinstance(ode, dict):
        ode_inputs = ode
    else:
        ode_inputs = get_promoted_vars(ode, iotypes=('input',))
    if user_targets is _unspecified:
        if name in ode_inputs:
            return [name]
    elif user_targets:
        if isinstance(user_targets, str):
            return [user_targets]
        else:
            return user_targets
    return []


def _configure_constraint_introspection(phase):
    """
    Modify constraint options in-place using introspection of the phase and its ODE.

    Parameters
    ----------
    phase : Phase
        The phase object whose boundary and path constraints are to be introspected.
    """
    from ..transcriptions import Birkhoff
    birkhoff = isinstance(phase.options['transcription'], Birkhoff)

    for constraint_type, constraints in [('initial', phase._initial_boundary_constraints),
                                         ('final', phase._final_boundary_constraints),
                                         ('path', phase._path_constraints)]:

        time_units = phase.time_options['units']
        time_name = phase.time_options['name']

        for con in constraints:
            # Determine the path to the variable which we will be constraining
            var = con['name']
            var_type = phase.classify_var(var)

            if var != con['constraint_name'] is not None and var_type != 'ode':
                om.issue_warning(f"Option 'constraint_name' on {constraint_type} constraint {var} is only "
                                 f"valid for ODE outputs. The option is being ignored.", om.UnusedOptionWarning)

            if var_type == 't':
                con['shape'] = (1,)
                con['units'] = time_units if con['units'] is None else con['units']
                con['constraint_path'] = f'timeseries.{time_name}'

            elif var_type == 't_phase':
                con['shape'] = (1,)
                con['units'] = time_units if con['units'] is None else con['units']
                con['constraint_path'] = f'timeseries.{time_name}_phase'

            elif var_type == 'state':
                prefix = 'states:' if phase.timeseries_options['use_prefix'] else ''
                state_shape = phase.state_options[var]['shape']
                state_units = phase.state_options[var]['units']
                con['shape'] = state_shape
                con['units'] = state_units if con['units'] is None else con['units']
                if birkhoff and constraint_type in ('initial', 'final'):
                    con['constraint_path'] = f'boundary_vals.{var}'
                else:
                    con['constraint_path'] = f'timeseries.{prefix}{var}'

            elif var_type == 'parameter':
                param_shape = phase.parameter_options[var]['shape']
                param_units = phase.parameter_options[var]['units']
                con['shape'] = param_shape
                con['units'] = param_units if con['units'] is None else con['units']
                con['constraint_path'] = f'parameter_vals:{var}'

            elif var_type == 'control':
                prefix = 'controls:' if phase.timeseries_options['use_prefix'] else ''
                control_shape = phase.control_options[var]['shape']
                control_units = phase.control_options[var]['units']

                con['shape'] = control_shape
                con['units'] = control_units if con['units'] is None else con['units']
                if birkhoff and constraint_type in ('initial', 'final'):
                    con['constraint_path'] = f'boundary_vals.{var}'
                else:
                    con['constraint_path'] = f'timeseries.{prefix}{var}'

            elif var_type == 'control_rate':
                prefix = 'control_rates:' if phase.timeseries_options['use_prefix'] else ''
                control_name = var[:-5]
                control_shape = phase.control_options[control_name]['shape']
                control_units = phase.control_options[control_name]['units']
                con['shape'] = control_shape
                con['units'] = get_rate_units(control_units, time_units, deriv=1) \
                    if con['units'] is None else con['units']
                if birkhoff and constraint_type in ('initial', 'final'):
                    con['constraint_path'] = f'boundary_vals.{var}'
                else:
                    con['constraint_path'] = f'timeseries.{prefix}{var}'

            elif var_type == 'control_rate2':
                prefix = 'control_rates:' if phase.timeseries_options['use_prefix'] else ''
                control_name = var[:-6]
                control_shape = phase.control_options[control_name]['shape']
                control_units = phase.control_options[control_name]['units']
                con['shape'] = control_shape
                con['units'] = get_rate_units(control_units, time_units, deriv=2) \
                    if con['units'] is None else con['units']
                if birkhoff and constraint_type in ('initial', 'final'):
                    con['constraint_path'] = f'boundary_vals.{var}'
                else:
                    con['constraint_path'] = f'timeseries.{prefix}{var}'
            else:
                # Failed to find variable, assume it is in the ODE. This requires introspection.
                ode = phase.options['transcription']._get_ode(phase)

                meta = get_source_metadata(ode, src=var, user_units=con['units'], user_shape=con['shape'])

                con['shape'] = meta['shape']
                con['units'] = meta['units']

                if birkhoff and constraint_type in ('initial', 'final'):
                    con['constraint_path'] = f'boundary_vals.{var}'
                else:
                    con['constraint_path'] = f'timeseries.{con["constraint_name"]}'


def configure_controls_introspection(control_options, ode, time_units='s'):
    """
    Modify control options in-place using introspection of the user-provided ODE.

    Parameters
    ----------
    control_options : dict of {str: ControlOptionsDictionary
        A dictionary keyed by control name containing the options for all controls to be applied
        to the ODE.
    ode : om.System
        An instantiated System that serves as the ODE to which the controls should be applied.
    time_units : str
        The units of time for the Phase.

    Raises
    ------
    ValueError
        If the control or one of its rates are connected to a variable that is tagged as static
        within the ODE.
    """
    ode_inputs = get_promoted_vars(ode, iotypes='input', metadata_keys=['shape', 'units', 'val', 'tags'])
    for name, options in control_options.items():
        targets = _get_targets_metadata(ode_inputs, name=name, user_targets=options['targets'])

        options['targets'] = list(targets.keys())
        if targets:
            if options['units'] is _unspecified:
                options['units'] = _get_common_metadata(targets, metadata_key='units')

            if options['shape'] in _none_or_unspecified:
                shape = _get_common_metadata(targets, metadata_key='shape')
                if len(shape) == 1:
                    options['shape'] = (1,)
                else:
                    options['shape'] = shape[1:]

            if any(['dymos.static_target' in meta['tags'] for meta in targets.values()]):
                raise ValueError(f"Control '{name}' cannot be connected to its targets because one "
                                 f"or more targets are tagged with 'dymos.static_target'.")

        # Now check rate targets
        rate_targets = _get_targets_metadata(ode_inputs, name=f'{name}_rate',
                                             user_targets=options['rate_targets'])

        options['rate_targets'] = list(rate_targets.keys())
        if rate_targets:
            if options['units'] is _unspecified:
                rate_target_units = _get_common_metadata(rate_targets, metadata_key='units')
                options['units'] = time_units if rate_target_units is None else \
                    simplify_unit(f'{rate_target_units}*{time_units}')

            if options['shape'] in _none_or_unspecified:
                shape = _get_common_metadata(rate_targets, metadata_key='shape')
                if len(shape) == 1:
                    options['shape'] = (1,)
                else:
                    options['shape'] = shape[1:]

            if any(['dymos.static_target' in meta['tags'] for meta in rate_targets.values()]):
                raise ValueError(f"Control rate of '{name}' cannot be connected to its targets because one "
                                 f"or more targets are tagged with 'dymos.static_target'.")

        # Now check rate2 targets
        rate2_targets = _get_targets_metadata(ode_inputs, name=f'{name}_rate2',
                                              user_targets=options['rate2_targets'])

        options['rate2_targets'] = list(rate2_targets.keys())
        if rate2_targets:
            if options['units'] is _unspecified:
                rate2_target_units = _get_common_metadata(rate_targets, metadata_key='units')
                options['units'] = f'{time_units**2}' if rate2_target_units is None \
                    else simplify_unit(f'{rate2_target_units}*{time_units}**2')

            if options['shape'] in _none_or_unspecified:
                shape = _get_common_metadata(rate2_targets, metadata_key='shape')
                if len(shape) == 1:
                    options['shape'] = (1,)
                else:
                    options['shape'] = shape[1:]

            if any(['dymos.static_target' in meta['tags'] for meta in rate2_targets.values()]):
                raise ValueError(f"Control rate2 of '{name}' cannot be connected to its targets because one "
                                 f"or more targets are tagged with 'dymos.static_target'.")


def configure_parameters_introspection(parameter_options, ode):
    """
    Modify parameter options in-place using introspection of the user-provided ODE.

    Parameters
    ----------
    parameter_options : dict of {str: ParameterOptionsDictionary
        A dictionary keyed by parameter name containing the options for all parameters to be applied
        to the ODE. Options for 'targets', 'units', 'shape', and 'static_target' are modified in-place.
    ode : om.System
        An instantiated System that serves as the ODE to which the parameters should be applied.
    """
    for name, options in parameter_options.items():
        try:
            targets = _get_targets_metadata(ode, name=name, user_targets=options['targets'])
        except ValueError as e:
            raise ValueError(f'Parameter `{name}` has invalid target(s).\n{str(e)}') from e

        options['targets'] = list(targets.keys())

        static_tagged_targets = {tgt for tgt, meta in targets.items() if 'dymos.static_target' in meta['tags']}

        # This is a bit of a hack. Any target with a shape of (1,) is unambiguously static.
        # We may want to consider forcing users to tag these as static for dymos 2.0.0
        shape_1_targets = {tgt for tgt, meta in targets.items() if meta['shape'] == (1,)}
        if options['static_targets'] is _unspecified:
            options['static_targets'] = static_tagged_targets.union(shape_1_targets)
        elif options['static_targets']:
            options['static_targets'] = options['targets'].copy()
        else:
            options['static_targets'] = []

        if static_tagged_targets and not options['static_targets']:
            raise ValueError(f"Parameter `{name}` has invalid target(s).\n"
                             f"User has specified 'static_target = False' for parameter `{name}`,\nbut one or more "
                             f"targets is tagged with 'dymos.static_target':\n{static_tagged_targets}")

        if options['units'] is _unspecified:
            options['units'] = _get_common_metadata(targets, metadata_key='units')

        # Check that all targets have the same shape.
        tgt_shapes = {}
        # First find the shapes of the static targets
        for tgt, meta in targets.items():
            if tgt in options['static_targets']:
                tgt_shapes[tgt] = meta['shape']
            else:
                if len(meta['shape']) == 1:
                    tgt_shapes[tgt] = (1,)
                else:
                    tgt_shapes[tgt] = meta['shape'][1:]
        # Check that they're unique
        if len(set(tgt_shapes.values())) > 1:
            raise RuntimeError(f'Invalid targets for parameter `{name}`.\n'
                               f'Targets have multiple shapes.\n'
                               f'{tgt_shapes}')
        elif len(set(tgt_shapes.values())) == 1:
            introspected_shape = next(iter(set(tgt_shapes.values())))
        else:
            introspected_shape = None

        if options['shape'] in _none_or_unspecified:
            if isinstance(options['val'], Number):
                options['shape'] = introspected_shape
            else:
                options['shape'] = np.asarray(options['val']).shape
        else:
            if introspected_shape is not None and options['shape'] != introspected_shape:
                raise RuntimeError(f'Shape provided to parameter `{name}` differs from its targets.\n'
                                   f'Given shape: {options["shape"]}\n'
                                   f'Target shapes:\n'
                                   f'{tgt_shapes}')

        options['val'], options['shape'] = ensure_compatible(name, options['val'], options['shape'])


def configure_time_introspection(time_options, ode):
    """
    Modify time options in-place using introspection of the user-provided ODE.

    Parameters
    ----------
    time_options : dict of {str: TimeOptionsDictionary
        A dictionary keyed by control name containing the options for all controls to be applied
        to the ODE. Options for 'targets', 't_phase_targets', and 'units' are modified in-place.
    ode : om.System
        An instantiated System that serves as the ODE to which the controls should be applied.

    Raises
    ------
    ValueError
        If time or time_phase are connected to a variable that is tagged as static
        within the ODE.
    """
    ode_inputs = get_promoted_vars(ode, 'input', metadata_keys=['shape', 'val', 'units', 'tags'])
    time_name = time_options['name']
    t_phase_name = f'{time_name}_phase'

    targets = _get_targets_metadata(ode_inputs, name=time_name, user_targets=time_options['targets'])

    time_options['targets'] = targets

    if time_options['units'] is _unspecified:
        time_options['units'] = _get_common_metadata(targets, 'units')

    if any(['dymos.static_target' in meta['tags'] for meta in targets.values()]):
        raise ValueError(f"The integration variable {time_name} cannot be connected to its targets because one "
                         f"or more targets are tagged with 'dymos.static_target'.")

    # t_phase
    targets = _get_targets_metadata(ode_inputs, name=t_phase_name, user_targets=time_options['time_phase_targets'])

    time_options['time_phase_targets'] = targets

    if any(['dymos.static_target' in meta['tags'] for meta in targets.values()]):
        raise ValueError("'t_phase' cannot be connected to its targets because one "
                         "or more targets are tagged with 'dymos.static_target'.")


def configure_states_introspection(state_options, time_options, control_options, parameter_options,
                                   ode):
    """
    Modifies state options in-place, automatically determining 'targets', 'units', and 'shape' if necessary.

    The precedence rules for the state shape and units are as follows:
    1. If the user has specified units and shape in the state options, use those.
    2a. If the user has not specified shape, and targets exist, then pull the shape from the targets.
    2b. If the user has not specified shape and no targets exist, then pull the shape from the rate source.
    2c. If shape cannot be inferred, assume (1,)
    3a. If the user has not specified units, first try to pull units from a target
    3b. If there are no targets, pull units from the rate source and multiply by time units.

    Parameters
    ----------
    state_options : dict of {str: StateOptionsDictionary}
        The state variables to be configured.
    time_options : TimeOptionsDictionary
        The time options.
    control_options : dict of {str: ControlOptionsDictionary}
        The options for each control.
    parameter_options : dict of {str: ParameterOptionsDictionary}
        The options for each parameter.
    ode : System
        The OpenMDAO system which provides the state rates as outputs.
    """
    time_units = time_options['units']
    ode_inputs = get_promoted_vars(ode, 'input', metadata_keys=['units', 'shape', 'val', 'tags'])
    ode_outputs = get_promoted_vars(ode, 'output', metadata_keys=['units', 'shape', 'val', 'tags'])

    for state_name, options in state_options.items():
        # Automatically determine targets of state if left _unspecified
        targets = _get_targets_metadata(ode_inputs, state_name,
                                        user_targets=options['targets'])

        options['targets'] = list(targets.keys())
        if targets:
            if options['units'] is _unspecified:
                options['units'] = _get_common_metadata(targets, metadata_key='units')

            if options['shape'] in _none_or_unspecified:
                shape = _get_common_metadata(targets, metadata_key='shape')
                if len(shape) == 1:
                    options['shape'] = (1,)
                else:
                    options['shape'] = shape[1:]

            if any(['dymos.static_target' in meta['tags'] for meta in targets.values()]):
                raise ValueError(f"State '{state_name}' cannot be connected to its targets because one "
                                 f"or more targets are tagged with 'dymos.static_target'.")

        # 3. Attempt rate-source introspection
        rate_src = options['rate_source']
        rate_src_type = classify_var(rate_src, time_options, state_options, parameter_options, control_options)

        if rate_src_type in {'t', 't_phase'}:
            rate_src_units = time_options['units']
            rate_src_shape = (1,)
        elif rate_src_type == 'state':
            rate_src_units = state_options[rate_src]['units']
            rate_src_shape = state_options[rate_src]['shape']
        elif rate_src_type == 'control':
            rate_src_units = control_options[rate_src]['units']
            rate_src_shape = control_options[rate_src]['shape']
        elif rate_src_type == 'parameter':
            rate_src_units = parameter_options[rate_src]['units']
            rate_src_shape = parameter_options[rate_src]['shape']
        elif rate_src_type == 'control_rate':
            control_name = rate_src[:-5]
            control = control_options[control_name]
            rate_src_units = get_rate_units(control['units'], time_units, deriv=1)
            rate_src_shape = control['shape']
        elif rate_src_type == 'control_rate2':
            control_name = rate_src[:-6]
            control = control_options[control_name]
            rate_src_units = get_rate_units(control['units'], time_units, deriv=2)
            rate_src_shape = control['shape']
        elif rate_src_type == 'ode':
            meta = get_source_metadata(ode_outputs, src=rate_src, user_units=options['units'], user_shape=options['shape'])
            rate_src_shape = meta['shape']
            rate_src_units = meta['units']
        else:
            rate_src_shape = (1,)
            rate_src_units = None

        if options['shape'] in _none_or_unspecified:
            options['shape'] = rate_src_shape

        if options['units'] is _unspecified:
            if rate_src_units is None:
                options['units'] = time_units
            else:
                if time_units is None:
                    raise RuntimeError(f'Unable to infer the units of state variable `{state_name}` from\n'
                                       f'the rate units because the time units of the phase are set to None.\n'
                                       f'Change the time units to something other than None, or explicitly\n'
                                       f'set the state units using one of the following options:\n'
                                       f'- Tag the state rate source `{rate_src}` with `dymos.state_units:{{units}}`\n'
                                       f'- Use the `set_state_options(\'{state_name}\', units={{units}})` method on the phase.')
                else:
                    options['units'] = simplify_unit(f'{rate_src_units}*{time_units}')


def configure_analytic_states_introspection(state_options, ode):
    """
    Modifies state options in-place, automatically determining 'targets', 'units', and 'shape' if necessary.

    The precedence rules for the state shape and units are as follows:
    1. If the user has specified units and shape in the state options, use those.
    2a. If the user has not specified shape, and targets exist, then pull the shape from the targets.
    2b. If the user has not specified shape and no targets exist, then pull the shape from the rate source.
    2c. If shape cannot be inferred, assume (1,)
    3a. If the user has not specified units, first try to pull units from a target
    3b. If there are no targets, pull units from the rate source and multiply by time units.

    Parameters
    ----------
    state_options : dict of {str: StateOptionsDictionary}
        The state variables to be configured.
    ode : System
        The OpenMDAO system which provides the state rates as outputs.
    """
    ode_outputs = get_promoted_vars(ode, 'output')

    for state_name, options in state_options.items():
        # Automatically determine targets of state if left _unspecified
        source = options['source'] if options['source'] else state_name
        meta = get_source_metadata(ode_outputs, src=source, user_units=options['units'], user_shape=options['shape'])
        src_shape = meta['shape']
        src_units = meta['units']

        if 'dymos.static_output' in meta['tags']:
            raise RuntimeError(f'ODE output {source} is tagged with `dymos.static_output` and cannot be used as a '
                               f'state variable in an AnalyticPhase.')

        if options['shape'] in _none_or_unspecified:
            options['shape'] = src_shape

        if options['units'] is _unspecified:
            options['units'] = src_units


def configure_states_discovery(state_options, ode):
    """
    Searches phase output metadata for any declared states and adds them.

    Parameters
    ----------
    state_options : dict
        The dictionary of options for each state in the phase.
    ode : System
        The System instance providing the ODE for the phase.
    """
    out_meta = ode.get_io_metadata(iotypes='output', metadata_keys=['tags'],
                                   get_remote=True)

    for name, meta in out_meta.items():
        tags = meta['tags']
        prom_name = meta['prom_name']
        state = None
        for tag in sorted(tags):

            # Declared as rate_source.
            if tag.startswith('dymos.state_rate_source:') or tag.startswith('state_rate_source:'):
                state = tag.rpartition(':')[-1]
                if tag.startswith('state_rate_source:'):
                    msg = f"The tag '{tag}' has a deprecated format and will no longer work in " \
                          f"dymos version 2.0.0. Use 'dymos.state_rate_source:{state}' instead."
                    om.issue_warning(msg, category=om.OMDeprecationWarning)
                if state not in state_options:
                    state_options[state] = StateOptionsDictionary()
                    state_options[state]['name'] = state

                if state_options[state]['rate_source'] is not None:
                    if state_options[state]['rate_source'] != prom_name:
                        raise ValueError(f"rate_source has been declared twice for state "
                                         f"'{state}' which is tagged on '{name}'.")

                state_options[state]['rate_source'] = prom_name

            # Declares units for state.
            if tag.startswith('dymos.state_units:') or tag.startswith('state_units:'):
                tagged_state_units = tag.rpartition(':')[-1]
                if tag.startswith('state_units:'):
                    msg = f"The tag '{tag}' has a deprecated format and will no longer work in " \
                          f"dymos version 2.0.0. Use 'dymos.{tag}' instead."
                    om.issue_warning(msg, category=om.OMDeprecationWarning)
                if state is None:
                    raise ValueError(f"'{tag}' tag declared on '{prom_name}' also requires "
                                     f"that the 'dymos.state_rate_source:{tagged_state_units}' "
                                     f"tag be declared.")
                state_options[state]['units'] = tagged_state_units

    # Check over all existing states and make sure we aren't missing any rate sources.
    for name, options in state_options.items():
        if options['rate_source'] is None:
            raise ValueError(f"State '{name}' is missing a rate_source.")


def configure_analytic_states_discovery(state_options, ode):
    """
    Searches phase output metadata for any declared states and adds them.

    Parameters
    ----------
    state_options : dict
        The dictionary of options for each state in the phase.
    ode : System
        The System instance providing the ODE for the phase.
    """
    out_meta = ode.get_io_metadata(iotypes='output', metadata_keys=['tags'],
                                   get_remote=True)

    for name, meta in out_meta.items():
        tags = meta['tags']
        prom_name = meta['prom_name']
        state = None
        for tag in sorted(tags):

            # Declared as rate_source.
            if tag.startswith('dymos.state_source:'):
                state = tag.rpartition(':')[-1]
                if state not in state_options:
                    state_options[state] = StateOptionsDictionary()
                    state_options[state]['name'] = state

                if state_options[state]['source'] is not None:
                    if state_options[state]['source'] != prom_name:
                        raise ValueError(f"source has been declared twice for state "
                                         f"'{state}' which is tagged on '{name}'.")

                state_options[state]['source'] = prom_name

            # Declares units for state.
            if tag.startswith('dymos.state_units:') or tag.startswith('state_units:'):
                tagged_state_units = tag.rpartition(':')[-1]
                if state is None:
                    raise ValueError(f"'{tag}' tag declared on '{prom_name}' also requires "
                                     f"that the 'dymos.state_source:{tagged_state_units}' "
                                     f"tag be declared.")
                state_options[state]['units'] = tagged_state_units

    # Check over all existing states and make sure we aren't missing any rate sources.
    for name, options in state_options.items():
        if options['source'] is None:
            raise ValueError(f"State '{name}' is missing a source.")


def configure_timeseries_output_glob_expansion(phase):
    """
    Modify timeseries outputs in-place using introspection to expand any glob patterns.

    Parameters
    ----------
    phase : Phase
        The phase object whose boundary and path constraints are to be introspected.
    """
    transcription = phase.options['transcription']
    ode = transcription._get_ode(phase)
    ode_outputs = get_promoted_vars(ode, 'output')

    new_outputs = {}

    # Step 1: Expand globs
    for ts_name, ts_meta in phase._timeseries.items():

        explicit_requests = set([output['name'] for output in ts_meta['outputs'].values() if '*' not in output['name']])

        for output_name, output_options in ts_meta['outputs'].items():

            if '*' in output_name:
                matching_outputs = filter_outputs(output_name, ode_outputs)
                wildcard_units = {} if output_options['wildcard_units'] is None else output_options['wildcard_units']

                for op, meta in matching_outputs.items():
                    if op not in explicit_requests and 'dymos.static_output' not in meta['tags']:
                        new_output = TimeseriesOutputOptionsDictionary()
                        new_output['name'] = op
                        new_output['output_name'] = opname = op.split('.')[-1]
                        new_output['units'] = _unspecified if op not in wildcard_units else wildcard_units[op]
                        new_output['shape'] = _unspecified
                        if opname not in explicit_requests and opname in new_outputs:
                            raise RuntimeError(f'{phase.pathname}: The glob pattern `{output_name}` matches multiple '
                                               f'outputs in the ODE.\n'
                                               f'Add these outputs explicitly with unique output names using the\n'
                                               f'output_name argument to avoid this error.\n'
                                               f'Colliding names: {op}  {new_outputs[opname]["name"]}')
                        else:
                            new_outputs[new_output['output_name']] = new_output
            else:
                new_outputs[output_name] = output_options

        phase._timeseries[ts_name]['outputs'] = new_outputs


def configure_timeseries_output_introspection(phase):
    """
    Modify timeseries outputs in-place using introspection to find output units and shape.

    Parameters
    ----------
    phase : Phase
        The phase object whose timeseries outputs are to be introspected.
    """
    configure_timeseries_output_glob_expansion(phase)

    transcription = phase.options['transcription']

    for ts_name, ts_opts in phase._timeseries.items():

        not_found = set()

        for output_name, output_options in ts_opts['outputs'].items():

            try:
                output_meta = transcription._get_timeseries_var_source(output_options['name'],
                                                                       output_options['output_name'],
                                                                       phase=phase)
            except ValueError:
                not_found.add(output_name)
                raise

            output_options['src'] = output_meta['src']
            output_options['src_idxs'] = output_meta['src_idxs']

            if output_options['shape'] in (None, _unspecified):
                output_options['shape'] = output_meta['shape']

            if output_options['units'] in (None, _unspecified):
                output_options['units'] = output_meta['units']

        if not_found:
            sorted_list = ', '.join(sorted([ts_opts['outputs'][output_name]['name'] for output_name in not_found]))
            om.issue_warning(f'{phase.pathname}: The following timeseries outputs were requested but not found in the '
                             f'ODE: {sorted_list}')

        for s in not_found:
            ts_opts['outputs'].pop(s)


def filter_outputs(patterns, sys):
    """
    Find all outputs of the given system that match one or more of the strings given in patterns.

    Parameters
    ----------
    patterns : str or Sequence
        A string or sequence of strings to be matched in the outputs of the given system.  These
        may include glob patterns.
    sys : System or dict
        The OpenMDAO system whose outputs are to be filtered, or a dictionary of outputs as returned by
        get_promoted_vars.

    Returns
    -------
    dict of {str: dict}
        A dictionary where the matching output names are the keys and the associated dict provides
        the 'units' and 'shapes' metadata.
    """
    outputs = sys if isinstance(sys, dict) else get_promoted_vars(sys, iotypes='output', metadata_keys=['shape', 'units', 'tags'])
    _patterns = [patterns] if isinstance(patterns, str) else patterns

    output_names = list(outputs.keys())
    filtered = set()
    results = {}

    for pattern in _patterns:
        filtered.update(fnmatch.filter(output_names, pattern))

    for var in filtered:
        results[var] = {'units': outputs[var]['units'], 'shape': outputs[var]['shape'], 'tags': outputs[var]['tags']}

    return results


def _configure_boundary_balance_introspection(phase):
    """
    Modify duration balance options in-place using introspection of the phase and its ODE.

    Parameters
    ----------
    phase : Phase
        The phase object whose boundary and path constraints are to be introspected.
    """
    time_units = phase.time_options['units']

    for param_name, options in phase.boundary_balance_options.items():
        resid_name = options['name']
        resid_units = options.get('eq_units', _unspecified)
        resid_shape = options.get('shape', None)

        # Determine the path to the variable which we will be constraining
        resid_type = phase.classify_var(resid_name)

        if param_name == 't_initial':
            param_units = phase.time_options['units']
            param_bounds = phase.time_options['initial_bounds']
        elif param_name == 't_duration':
            param_units = phase.time_options['units']
            param_bounds = phase.time_options['duration_bounds']
        elif param_name in phase.parameter_options:
            param_units = phase.parameter_options[param_name]['units']
            param_bounds = tuple(phase.parameter_options[param_name][k] for k in ('lower', 'upper'))
        elif param_name.startswith('initial_states:'):
            state_name = ':'.join(param_name.split(':')[1:])
            param_units = phase.state_options[state_name]['units']
            param_bounds = phase.state_options[state_name]['initial_bounds']
        elif param_name.startswith('final_states:'):
            state_name = ':'.join(param_name.split(':')[1:])
            param_units = phase.state_options[state_name]['units']
            param_bounds = phase.state_options[state_name]['final_bounds']
        else:
            raise ValueError(f'{phase.msginfo}: For boundary balance, param must be one of t_initial, t_duration, '
                             'a parameter in the phase, initial_states:{name}, or final_states:{name}')

        if options.get('units', _unspecified) is _unspecified:
            options['units'] = param_units

        if options.get('lower', _unspecified) is _unspecified:
            if param_bounds is None:
                options['lower'] = None
            else:
                options['lower'] = param_bounds[0]

        if options.get('upper', _unspecified) is _unspecified:
            if param_bounds is None:
                options['upper'] = None
            else:
                options['upper'] = param_bounds[1]

        if resid_units is _unspecified:
            if resid_type in ['t', 't_phase']:
                options['eq_units'] = time_units
            elif resid_type == 'state':
                if resid_name.startswith('initial_states:') or resid_name.startswith('final_states:'):
                    state_name = ':'.join(resid_name.split(':')[1:])
                else:
                    state_name = resid_name
                options['eq_units'] = phase.state_options[state_name]['units']
            elif resid_type == 'parameter':
                options['eq_units'] = phase.parameter[resid_name]['units']
            elif resid_type == 'control':
                options['eq_units'] = phase.control_options[resid_name]['units']
            elif resid_type == 'control_rate':
                control_units = phase.state_options[resid_name]['units']
                options['eq_units'] = get_rate_units(control_units, time_units, deriv=1)
            elif resid_type == 'control_rate2':
                options['eq_units'] = get_rate_units(control_units, time_units, deriv=2)
            elif resid_type == 'ode':
                # Failed to find variable, assume it is in the ODE. This requires introspection.
                ode = phase.options['transcription']._get_ode(phase)
                meta = get_source_metadata(ode, src=resid_name, user_units=resid_units, user_shape=resid_shape)
                options['eq_units'] = meta['units']
            else:
                raise ValueError(f'{phase.msginfo}: Unable to find boundary balance name {resid_name}')


def _configure_constraint_introspection(phase):
    """
    Modify constraint options in-place using introspection of the phase and its ODE.

    Parameters
    ----------
    phase : Phase
        The phase object whose boundary and path constraints are to be introspected.
    """
    from ..transcriptions import Birkhoff
    birkhoff = isinstance(phase.options['transcription'], Birkhoff)

    for constraint_type, constraints in [('initial', phase._initial_boundary_constraints),
                                         ('final', phase._final_boundary_constraints),
                                         ('path', phase._path_constraints)]:

        time_units = phase.time_options['units']
        time_name = phase.time_options['name']

        for con in constraints:
            # Determine the path to the variable which we will be constraining
            var = con['name']
            var_type = phase.classify_var(var)

            if var != con['constraint_name'] is not None and var_type != 'ode':
                om.issue_warning(f"Option 'constraint_name' on {constraint_type} constraint {var} is only "
                                 f"valid for ODE outputs. The option is being ignored.", om.UnusedOptionWarning)

            if var_type == 't':
                con['shape'] = (1,)
                con['units'] = time_units if con['units'] is None else con['units']
                con['constraint_path'] = f'timeseries.{time_name}'

            elif var_type == 't_phase':
                con['shape'] = (1,)
                con['units'] = time_units if con['units'] is None else con['units']
                con['constraint_path'] = f'timeseries.{time_name}_phase'

            elif var_type == 'state':
                prefix = 'states:' if phase.timeseries_options['use_prefix'] else ''
                state_shape = phase.state_options[var]['shape']
                state_units = phase.state_options[var]['units']
                con['shape'] = state_shape
                con['units'] = state_units if con['units'] is None else con['units']
                if birkhoff and constraint_type in ('initial', 'final'):
                    con['constraint_path'] = f'boundary_vals.{var}'
                else:
                    con['constraint_path'] = f'timeseries.{prefix}{var}'

            elif var_type == 'parameter':
                param_shape = phase.parameter_options[var]['shape']
                param_units = phase.parameter_options[var]['units']
                con['shape'] = param_shape
                con['units'] = param_units if con['units'] is None else con['units']
                con['constraint_path'] = f'parameter_vals:{var}'

            elif var_type == 'control':
                prefix = 'controls:' if phase.timeseries_options['use_prefix'] else ''
                control_shape = phase.control_options[var]['shape']
                control_units = phase.control_options[var]['units']

                con['shape'] = control_shape
                con['units'] = control_units if con['units'] is None else con['units']
                if birkhoff and constraint_type in ('initial', 'final'):
                    con['constraint_path'] = f'boundary_vals.{var}'
                else:
                    con['constraint_path'] = f'timeseries.{prefix}{var}'

            elif var_type == 'control_rate':
                prefix = 'control_rates:' if phase.timeseries_options['use_prefix'] else ''
                control_name = var[:-5]
                control_shape = phase.control_options[control_name]['shape']
                control_units = phase.control_options[control_name]['units']
                con['shape'] = control_shape
                con['units'] = get_rate_units(control_units, time_units, deriv=1) \
                    if con['units'] is None else con['units']
                if birkhoff and constraint_type in ('initial', 'final'):
                    con['constraint_path'] = f'boundary_vals.{var}'
                else:
                    con['constraint_path'] = f'timeseries.{prefix}{var}'

            elif var_type == 'control_rate2':
                prefix = 'control_rates:' if phase.timeseries_options['use_prefix'] else ''
                control_name = var[:-6]
                control_shape = phase.control_options[control_name]['shape']
                control_units = phase.control_options[control_name]['units']
                con['shape'] = control_shape
                con['units'] = get_rate_units(control_units, time_units, deriv=2) \
                    if con['units'] is None else con['units']
                if birkhoff and constraint_type in ('initial', 'final'):
                    con['constraint_path'] = f'boundary_vals.{var}'
                else:
                    con['constraint_path'] = f'timeseries.{prefix}{var}'

            else:
                # Failed to find variable, assume it is in the ODE. This requires introspection.
                ode = phase.options['transcription']._get_ode(phase)

                meta = get_source_metadata(ode, src=var, user_units=con['units'], user_shape=con['shape'])

                con['shape'] = meta['shape']
                con['units'] = meta['units']

                if birkhoff and constraint_type in ('initial', 'final'):
                    con['constraint_path'] = f'boundary_vals.{var}'
                else:
                    con['constraint_path'] = f'timeseries.{con["constraint_name"]}'


def _get_targets_metadata(ode, name, user_targets=_unspecified):
    """
    Return the targets of a variable in a given ODE system and their metadata.

    If the targets of the state is _unspecified, and the state name is a top level input name
    in the ODE, then the state values are automatically connected to that top-level input.
    If _unspecified and not a top-level input of the ODE, no connection is made.
    If targets is explicitly None, then no connection is made.
    Otherwise, if the user specified some other string or sequence of strings as targets, then
    those are returned.

    Parameters
    ----------
    ode : om.System or dict
        The OpenMDAO system which serves as the ODE for dymos, or a dictionary of the ODE inputs and their metadata
        from a previous call to _get_targets_metadata. If given as the ODE system, it should already have had its setup
        and configure methods called.
    name : str
        The name of the variable whose targets are desired.
    user_targets : str or None or Sequence or _unspecified
        Targets for the variable as given by the user.

    Returns
    -------
    targets : dict
        A dictionary of target inputs in the ODE and metadata associated with each target.

    Notes
    -----
    This method requires that the ODE has run its setup and configure methods.  Thus,
    this method should be called from configure of some parent Group, and the ODE should
    be a system within that Group.
    """
    ode_inputs = ode if isinstance(ode, dict) else get_promoted_vars(ode,
                                                                     iotypes='input',
                                                                     metadata_keys=['shape', 'units',
                                                                                    'val', 'tags'])

    targets = {t: {} for t in get_targets(ode_inputs, name, user_targets=user_targets)}

    for tgt in targets:
        if tgt not in ode_inputs:
            raise ValueError(f"No such ODE input: '{tgt}'.")

    for tgt, options in targets.items():
        options['units'] = ode_inputs[tgt]['units']
        options['shape'] = ode_inputs[tgt]['shape']
        options['val'] = ode_inputs[tgt]['val']
        options['tags'] = ode_inputs[tgt]['tags']

    return targets


def _get_common_metadata(targets, metadata_key):
    """
    Given a dictionary containing targets and their metadata, return the value associated
    with the given metadata key if that value is common to all targets, otherwise raise an
    Exception.

    Parameters
    ----------
    targets : dict
        A dictionary of targets and their metadata which must include the desired metadata key.
    metadata_key : str
        The metadata key desired.

    Returns
    -------
    object
        The common metadata value shared by all targets.

    Raises
    ------
    ValueError
        ValueError is raised if the targets do not all have the same metadata value.
    """
    meta_set = {meta[metadata_key] for meta in targets.values()}

    if len(meta_set) == 1:
        return next(iter(meta_set))
    elif len(meta_set) == 0:
        raise RuntimeError(f'Unable to automatically assign {metadata_key} based on targets. \n'
                           f'No targets were found.')
    else:
        err_dict = {tgt: meta[metadata_key] for tgt, meta in targets.items()}
        raise RuntimeError(f'Unable to automatically assign {metadata_key} based on targets.\n'
                           f'Targets have multiple {metadata_key} assigned:\n{err_dict}.\n'
                           f'Either promote targets and use set_input_defaults to assign common\n'
                           f'{metadata_key}, or explicitly provide {metadata_key} to the variable.')


def get_source_metadata(ode, src, user_units=_unspecified, user_shape=_unspecified):
    """
    Return the units and shape of output src in the given ODE.

    If the targets of the state is _unspecified, and the state name is a top level input name
    in the ODE, then the state values are automatically connected to that top-level input.
    If _unspecified and not a top-level input of the ODE, no connection is made.
    If targets is explicitly None, then no connection is made.
    Otherwise, if the user specified some other string or sequence of strings as targets, then
    those are returned.

    Parameters
    ----------
    ode : openmdao.core.System or dict
        The OpenMDAO system which serves as the ODE for dymos, or a dictionary of promoted output names and
        corresponding metadata from get_promoted_vars.  If given as a system, it should already have
        had its setup and configure methods called.
    src : str
        The relative path in the ODE to the source variable whose metadata is requested.
    user_units : str or None or Sequence or _unspecified
        Units for the variable as given by the user.
    user_shape : str or None or Sequence or _unspecified
        Shape for the variable as given by the user.

    Returns
    -------
    dict
        A dictionary containing the metadata for the source. This consists of shape, units, and tags.

    Notes
    -----
    This method requires that the ODE has run its setup and configure methods.  Thus,
    this method should be called from configure of some parent Group, and the ODE should
    be a system within that Group.
    """
    meta = {}
    ode_outputs = ode if isinstance(ode, dict) else get_promoted_vars(ode, iotypes='output')

    if src not in ode_outputs:
        raise ValueError(f"Unable to find the source '{src}' in the ODE.")

    if user_units in _none_or_unspecified:
        meta['units'] = ode_outputs[src]['units']
    else:
        meta['units'] = user_units

    if user_shape in _none_or_unspecified:
        ode_shape = ode_outputs[src]['shape']
        meta['shape'] = (1,) if len(ode_shape) == 1 else ode_shape[1:]
    else:
        meta['shape'] = user_shape

    meta['tags'] = ode_outputs[src]['tags']

    return meta



import unittest
import warnings

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse
from openmdao.utils.mpi import MPI, multi_proc_exception_check

from dymos.examples.finite_burn_orbit_raise.finite_burn_orbit_raise_problem import two_burn_orbit_raise_problem
from dymos.utils.misc import om_version
from dymos.utils.testing_utils import assert_cases_equal


@require_pyoptsparse(optimizer='IPOPT')
@use_tempdirs
class TestExampleTwoBurnOrbitRaiseConnectedRestart(unittest.TestCase):

    N_PROCS = 3

    @unittest.skipUnless(MPI, "MPI is required.")
    def test_ex_two_burn_orbit_raise_connected(self):
        optimizer = 'IPOPT'

        p = two_burn_orbit_raise_problem(transcription='gauss-lobatto', transcription_order=3,
                                         compressed=False, optimizer=optimizer,
                                         show_output=False, connected=True)

        with multi_proc_exception_check(p.comm):
            if p.model.traj.phases.burn2 in p.model.traj.phases._subsystems_myproc:
                assert_near_equal(p.get_val('traj.burn2.states:deltav')[0], 0.3995,
                                  tolerance=4.0E-3)

        sol_db = 'dymos_solution.db'
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        sim_case1 = om.CaseReader(sim_db).get_case('final')

        # Run again without an actual optimizer
        p = two_burn_orbit_raise_problem(transcription='gauss-lobatto', transcription_order=3,
                                         compressed=False, optimizer=optimizer,
                                         show_output=False, connected=True, run_driver=False,
                                         restart=sol_db)

        p.run_model()

        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        sim_case2 = om.CaseReader(sim_db).get_case('final')

        # Verify that the second case has the same inputs and outputs
        assert_cases_equal(sim_case1, sim_case2, tol=1.0E-7)

    def test_restart_from_solution_radau(self):
        optimizer = 'IPOPT'

        p = two_burn_orbit_raise_problem(transcription='radau', transcription_order=3,
                                         compressed=False, optimizer=optimizer, show_output=False)

        sol_db = 'dymos_solution.db'
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        sim_case1 = om.CaseReader(sim_db).get_case('final')

        with multi_proc_exception_check(p.comm):
            if p.model.traj.phases.burn2 in p.model.traj.phases._subsystems_myproc:
                assert_near_equal(p.get_val('traj.burn2.states:deltav')[-1], 0.3995,
                                  tolerance=2.0E-3)

        # Run again without an actual optimzier
        two_burn_orbit_raise_problem(transcription='radau', transcription_order=3,
                                     compressed=False, optimizer=optimizer, run_driver=False,
                                     show_output=False, restart=sol_db)

        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        sim_case2 = om.CaseReader(sim_db).get_case('final')

        # Verify that the second case has the same inputs and outputs
        assert_cases_equal(sim_case1, sim_case2, tol=1.0E-7)


@require_pyoptsparse(optimizer='IPOPT')
@use_tempdirs
class TestExampleTwoBurnOrbitRaiseConnected(unittest.TestCase):

    N_PROCS = 3

    @unittest.skipUnless(MPI, "MPI is required.")
    def test_ex_two_burn_orbit_raise_connected(self):
        optimizer = 'IPOPT'

        unexpected_warnings = \
            [(om.OpenMDAOWarning,
              "'traj' <class Trajectory>: Setting phases.nonlinear_solver to `om.NonlinearBlockJac(iprint=0)`.\n"
              "Connected phases in parallel require a non-default nonlinear solver.\n"
              "Use traj.options[\'default_nonlinear_solver\'] to explicitly set the solver."),
             (om.OpenMDAOWarning,
              "'traj' <class Trajectory>: Setting phases.linear_solver to `om.PETScKrylov()`.\n"
              "Connected phases in parallel require a non-default linear solver.\n"
              "Use traj.options[\'default_linear_solver\'] to explicitly set the solver.")]

        with warnings.catch_warnings(record=True) as w:
            p = two_burn_orbit_raise_problem(transcription='gauss-lobatto', transcription_order=3,
                                             compressed=False, optimizer=optimizer,
                                             show_output=False, connected=True,
                                             default_nonlinear_solver=om.NonlinearBlockJac(iprint=0),
                                             default_linear_solver=om.PETScKrylov())

        for category, msg in unexpected_warnings:
            for warn in w:
                if (issubclass(warn.category, category) and str(warn.message) == msg):
                    raise AssertionError(f"Saw unexpected warning {category.__name__}: {msg}")

        with multi_proc_exception_check(p.comm):
            if p.model.traj.phases.burn2 in p.model.traj.phases._subsystems_myproc:
                assert_near_equal(p.get_val('traj.burn2.states:deltav')[0], 0.3995,
                                  tolerance=4.0E-3)

        sol_db = 'dymos_solution.db'
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        case1 = om.CaseReader(sol_db).get_case('final')
        sim_case1 = om.CaseReader(sim_db).get_case('final')

        # Run again without an actual optimizer
        p = two_burn_orbit_raise_problem(transcription='gauss-lobatto', transcription_order=3,
                                         compressed=False, optimizer=optimizer, run_driver=False,
                                         show_output=False, restart=sol_db,
                                         connected=True, solution_record_file='dymos_solution2.db',
                                         simulation_record_file='dymos_simulation2.db')

        sol_db = 'dymos_solution2.db'
        sim_db = 'dymos_simulation2.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        case2 = om.CaseReader(sol_db).get_case('final')
        sim_case2 = om.CaseReader(sim_db).get_case('final')

        # Verify that the second case has the same inputs and outputs
        assert_cases_equal(case1, case2, tol=1.0E-7)
        assert_cases_equal(sim_case1, sim_case2, tol=1.0E-7)

    @unittest.skipUnless(MPI, "MPI is required.")
    def test_restart_from_solution_radau_to_connected(self):
        optimizer = 'IPOPT'

        p = two_burn_orbit_raise_problem(transcription='radau', transcription_order=3,
                                         compressed=False, optimizer=optimizer,
                                         show_output=False, connected=True)

        with multi_proc_exception_check(p.comm):
            if p.model.traj.phases.burn2 in p.model.traj.phases._subsystems_myproc:
                assert_near_equal(p.get_val('traj.burn2.states:deltav')[0], 0.3995,
                                  tolerance=4.0E-3)

        sol_db = 'dymos_solution.db'
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        case1 = om.CaseReader(sol_db).get_case('final')
        sim_case1 = om.CaseReader(sim_db).get_case('final')

        # Run again without an actual optimizer
        p2 = two_burn_orbit_raise_problem(transcription='radau', transcription_order=3,
                                          compressed=False, optimizer=optimizer, run_driver=False,
                                          show_output=False, restart=sol_db,
                                          connected=True, solution_record_file='dymos_solution2.db',
                                          simulation_record_file='dymos_simulation2.db')

        sol_db = 'dymos_solution2.db'
        sim_db = 'dymos_simulation2.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p2.get_outputs_dir() / sol_db
            sim_db = p2.model.traj.sim_prob.get_outputs_dir() / sim_db

        case2 = om.CaseReader(sol_db).get_case('final')
        sim_case2 = om.CaseReader(sim_db).get_case('final')

        # Verify that the second case has the same inputs and outputs
        assert_cases_equal(case1, case2, tol=1.0E-7)
        assert_cases_equal(sim_case1, sim_case2, tol=1.0E-7)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

import dymos as dm
from dymos.utils.lgl import lgl
from dymos.utils.misc import om_version
from dymos.examples.aircraft_steady_flight.aircraft_ode import AircraftODE


def ex_aircraft_steady_flight(transcription, optimizer='SLSQP', use_boundary_constraints=False,
                              show_output=True):
    p = om.Problem(model=om.Group())
    p.driver = om.pyOptSparseDriver()
    p.driver.options['optimizer'] = optimizer
    p.driver.declare_coloring()
    if optimizer == 'SNOPT':
        p.driver.opt_settings['Major iterations limit'] = 50
        p.driver.opt_settings['Major feasibility tolerance'] = 1.0E-6
        p.driver.opt_settings['Major optimality tolerance'] = 1.0E-6
        p.driver.opt_settings['Linesearch tolerance'] = 0.10
        if show_output:
            p.driver.opt_settings['iSumm'] = 6
    elif optimizer == 'IPOPT':
        p.driver.opt_settings['mu_init'] = 1e-3
        p.driver.opt_settings['max_iter'] = 200
        p.driver.opt_settings['acceptable_tol'] = 1e-6
        p.driver.opt_settings['constr_viol_tol'] = 1e-6
        p.driver.opt_settings['compl_inf_tol'] = 1e-6
        p.driver.opt_settings['acceptable_iter'] = 0
        p.driver.opt_settings['tol'] = 1e-6
        p.driver.opt_settings['nlp_scaling_method'] = 'gradient-based'  # for faster convergence
        p.driver.opt_settings['alpha_for_y'] = 'safer-min-dual-infeas'
        p.driver.opt_settings['mu_strategy'] = 'monotone'
        # p.driver.opt_settings['derivative_test'] = 'first-order'
        p.driver.opt_settings['print_level'] = 0
    elif optimizer == 'SLSQP':
        p.driver.opt_settings['MAXIT'] = 100

    phase = dm.Phase(ode_class=AircraftODE,
                     transcription=transcription)

    # Pass Reference Area from an external source
    assumptions = p.model.add_subsystem('assumptions', om.IndepVarComp())
    assumptions.add_output('S', val=427.8, units='m**2')
    assumptions.add_output('mass_empty', val=1.0, units='kg')
    assumptions.add_output('mass_payload', val=1.0, units='kg')

    traj = p.model.add_subsystem('traj', dm.Trajectory())
    traj.add_phase('phase0', phase)

    phase.set_time_options(fix_initial=True,
                           duration_bounds=(300, 10000),
                           duration_ref=5600)

    fix_final = True
    if use_boundary_constraints:
        fix_final = False
        phase.add_boundary_constraint('mass_fuel', loc='final',
                                      equals=1e-3, linear=False)
        phase.add_boundary_constraint('alt', loc='final', equals=10.0, linear=False)

    phase.add_state('range', units='NM',
                    rate_source='range_rate_comp.dXdt:range',
                    fix_initial=True, fix_final=False, ref=1,
                    defect_ref=1, lower=0, upper=2000)

    phase.add_state('mass_fuel', units='lbm',
                    rate_source='propulsion.dXdt:mass_fuel',
                    fix_initial=True, fix_final=fix_final,
                    upper=1.5E5, lower=0.0, ref=1e4, defect_ref=1e4)

    phase.add_state('alt', units='kft',
                    rate_source='climb_rate',
                    fix_initial=True, fix_final=fix_final,
                    lower=0.0, upper=60, ref=1, defect_ref=1)

    phase.add_control('climb_rate', units='ft/min', opt=True, lower=-3000, upper=3000,
                      targets=['gam_comp.climb_rate'],
                      ref0=-10_000, ref=10_000,
                      rate_continuity=True, rate2_continuity=False)

    phase.add_control('mach', targets=['tas_comp.mach', 'aero.mach'], opt=False)

    phase.add_parameter('S',
                        targets=['aero.S', 'flight_equilibrium.S', 'propulsion.S'],
                        units='m**2')

    phase.add_parameter('mass_empty', targets=['mass_comp.mass_empty'], units='kg')
    phase.add_parameter('mass_payload', targets=['mass_comp.mass_payload'], units='kg')

    phase.add_path_constraint('propulsion.tau', lower=0.01, upper=2.0)

    p.model.connect('assumptions.S', 'traj.phase0.parameters:S')
    p.model.connect('assumptions.mass_empty', 'traj.phase0.parameters:mass_empty')
    p.model.connect('assumptions.mass_payload', 'traj.phase0.parameters:mass_payload')

    phase.add_objective('range', loc='final', ref=-100.0)

    p.setup()

    phase.set_time_val(initial=0.0, duration=5000.0)
    phase.set_state_val('range', (0, 724.0))
    phase.set_state_val('mass_fuel', (30000, 1e-3))
    phase.set_state_val('alt', 10.0)
    phase.set_control_val('mach', 0.8)

    p['assumptions.S'] = 427.8
    p['assumptions.mass_empty'] = 0.15E6
    p['assumptions.mass_payload'] = 84.02869 * 400

    dm.run_problem(p, simulate=False, make_plots=True)

    return p


@use_tempdirs
class TestExSteadyAircraftFlight(unittest.TestCase):

    @require_pyoptsparse(optimizer='SLSQP')
    def test_ex_aircraft_steady_flight_opt_radau(self):
        num_seg = 15
        seg_ends, _ = lgl(num_seg + 1)

        tx = dm.Radau(num_segments=num_seg, segment_ends=seg_ends, order=3, compressed=False)
        p = ex_aircraft_steady_flight(transcription=tx, optimizer='SLSQP')
        assert_near_equal(p.get_val('traj.phase0.timeseries.range', units='NM')[-1],
                          726.85, tolerance=1.0E-2)

    @require_pyoptsparse(optimizer='IPOPT')
    @unittest.skipIf(om_version()[0] < (3, 32, 2), 'Test requires OpenMDAO 3.32.2 or later')
    def test_ex_aircraft_steady_flight_opt_birkhoff(self):

        tx = dm.Birkhoff(num_nodes=50, grid_type='cgl',
                         solve_segments=False)
        p = ex_aircraft_steady_flight(transcription=tx, optimizer='IPOPT')
        assert_near_equal(p.get_val('traj.phase0.timeseries.range', units='NM')[-1],
                          726.85, tolerance=2.0E-2)

    @require_pyoptsparse(optimizer='SLSQP')
    def test_ex_aircraft_steady_flight_solve_radau(self):
        num_seg = 15
        seg_ends, _ = lgl(num_seg + 1)

        tx = dm.Radau(num_segments=num_seg, segment_ends=seg_ends, order=3, compressed=False,
                      solve_segments='forward')
        p = ex_aircraft_steady_flight(transcription=tx, optimizer='SLSQP',
                                      use_boundary_constraints=True)
        assert_near_equal(p.get_val('traj.phase0.timeseries.range', units='NM')[-1],
                          726.85, tolerance=1.0E-2)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import pathlib
import re
import subprocess
import tempfile
import unittest
from openmdao.utils.testing_utils import use_tempdirs
from openmdao.utils.assert_utils import assert_near_equal


def extract_python_code(filename):
    """Extracts Python code blocks from a Markdown file and save them to a temp file"""
    with open(filename, "r") as f:
        content = f.read()

    # Find triple-backtick Python code blocks
    code_blocks = re.findall(r"```python(.*?)```", content, re.DOTALL)

    # Create a temporary Python file
    temp_file = tempfile.NamedTemporaryFile(delete=False, suffix=".py", mode="w")
    temp_file.write('\n'.join(code_blocks))
    temp_file.close()

    return temp_file.name  # Return the filename


@use_tempdirs
class TestReadme(unittest.TestCase):

    def test_readme_code(self):
        """
        Test that the code in the readme works without exceptions.

        Run this test from a developer install location.
        If run from site-packages it will fail to find the readme and skip.
        """
        readme_path = pathlib.Path(__file__).parent.parent.parent / 'readme.md'
        try:
            script = extract_python_code(readme_path)
        except FileNotFoundError:
            self.skipTest('Test cannot be run from site-packages.')

        # Run the extracted Python file as a subprocess
        result = subprocess.run(["python", script], capture_output=True, text=True)

        # Print output for debugging
        # print("STDOUT:", result.stdout)
        # print("STDERR:", result.stderr)

        match = re.search(r'Current function value:\s+([-+]?\d*\.?\d+)', result.stdout)
        assert_near_equal(float(match.groups()[0]), 1.8016, tolerance=3)

        # Ensure the script runs without errors
        self.assertEqual(result.returncode, 0, f'Python script failed with errors:\n{result.stderr}')



import openmdao.api as om
import numpy as np


class LowThrustODE(om.ExplicitComponent):
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input('r', val=np.zeros(nn), units=None)
        self.add_input('theta', val=np.zeros(nn), units='rad')
        self.add_input('vr', val=np.zeros(nn), units='1/s')
        self.add_input('vt', val=np.zeros(nn), units='1/s')
        self.add_input('alpha', val=np.zeros(nn), units='rad')

        self.add_output('theta_dot', val=np.zeros(nn), units='rad/s')
        self.add_output('vr_dot', val=np.zeros(nn), units='1/s**2')
        self.add_output('vt_dot', val=np.zeros(nn), units='1/s**2')

        ar = np.arange(nn)

        self.declare_partials(of='theta_dot', wrt='r', rows=ar, cols=ar)
        self.declare_partials(of='theta_dot', wrt='vt', rows=ar, cols=ar)

        self.declare_partials(of='vr_dot', wrt='r', rows=ar, cols=ar)
        self.declare_partials(of='vr_dot', wrt='vt', rows=ar, cols=ar)
        self.declare_partials(of='vr_dot', wrt='alpha', rows=ar, cols=ar)

        self.declare_partials(of='vt_dot', wrt='r', rows=ar, cols=ar)
        self.declare_partials(of='vt_dot', wrt='vr', rows=ar, cols=ar)
        self.declare_partials(of='vt_dot', wrt='vt', rows=ar, cols=ar)
        self.declare_partials(of='vt_dot', wrt='alpha', rows=ar, cols=ar)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        r = inputs['r']
        vr = inputs['vr']
        vt = inputs['vt']
        alpha = inputs['alpha']

        T = 1e-3

        outputs['theta_dot'] = vt/r
        outputs['vr_dot'] = vt**2/r - 1/r**2 + T*np.sin(alpha)
        outputs['vt_dot'] = -vr*vt/r + T*np.cos(alpha)

    def compute_partials(self, inputs, partials, discrete_inputs=None):
        r = inputs['r']
        vr = inputs['vr']
        vt = inputs['vt']
        alpha = inputs['alpha']

        T = 1e-3

        partials['theta_dot', 'r'] = -vt/r**2
        partials['theta_dot', 'vt'] = 1/r

        partials['vr_dot', 'r'] = -(vt/r)**2 + 2/r**3
        partials['vr_dot', 'vt'] = 2*vt/r
        partials['vr_dot', 'alpha'] = T * np.cos(alpha)

        partials['vt_dot', 'r'] = vt*vr/r**2
        partials['vt_dot', 'vr'] = -vt/r
        partials['vt_dot', 'vt'] = -vr/r
        partials['vt_dot', 'alpha'] = -T * np.sin(alpha)



import unittest

import numpy as np
from numpy.testing import assert_almost_equal

import openmdao.api as om

from dymos.examples.min_time_climb.prop.max_thrust_comp import MaxThrustComp, THR_DATA


class TestBrysonThrustComp(unittest.TestCase):

    def test_grid_values(self):
        n = 10

        p = om.Problem(model=om.Group())

        ivc = om.IndepVarComp()
        ivc.add_output(name='h', val=np.zeros(n), units='ft')
        ivc.add_output(name='mach', val=np.zeros(n), units=None)

        p.model.add_subsystem(name='ivc', subsys=ivc, promotes_outputs=['h', 'mach'])
        p.model.add_subsystem(name='tcomp',
                              subsys=MaxThrustComp(vec_size=n, extrapolate=True, method='cubic'))

        p.model.connect('h', 'tcomp.h')
        p.model.connect('mach', 'tcomp.mach')

        p.setup()

        p['mach'] = THR_DATA['mach']
        for i in range(10):
            p['h'] = THR_DATA['h'][i] * np.ones(n)
            p.run_model()
            thrust_N = p['tcomp.max_thrust']
            assert_almost_equal(thrust_N, THR_DATA['thrust'][i, :])


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import os
import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal, assert_check_partials
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

import dymos as dm
from dymos.examples.double_integrator.breakwell_ode import BreakwellODE


@require_pyoptsparse(optimizer='IPOPT')
def double_integrator_direct_collocation(transcription='gauss-lobatto', compressed=True, optimizer='IPOPT'):

    p = om.Problem(model=om.Group())
    p.driver = om.pyOptSparseDriver()
    p.driver.declare_coloring()
    p.driver.options['optimizer'] = optimizer

    if optimizer == 'IPOPT':
        p.driver.opt_settings['max_iter'] = 5000
        p.driver.opt_settings['alpha_for_y'] = 'safer-min-dual-infeas'
        p.driver.opt_settings['print_level'] = 0
        p.driver.opt_settings['nlp_scaling_method'] = 'gradient-based'
        p.driver.opt_settings['tol'] = 1.0E-3
        p.driver.opt_settings['constr_viol_tol'] = 1.0E-6
    elif optimizer == 'SNOPT':
        p.driver.opt_settings['Major iterations limit'] = 1000
        p.driver.opt_settings['Major feasibility tolerance'] = 1.0E-6
        p.driver.opt_settings['Major optimality tolerance'] = 1.0E-3
        p.driver.opt_settings['iSumm'] = 6

    if transcription == 'gauss-lobatto':
        t = dm.GaussLobatto(num_segments=30, order=3, compressed=compressed)
    elif transcription == "radau-ps":
        t = dm.Radau(num_segments=30, order=3, compressed=compressed)
    elif transcription == 'birkhoff':
        t = dm.Birkhoff(num_nodes=51)
    else:
        raise ValueError('invalid transcription')

    traj = p.model.add_subsystem('traj', dm.Trajectory())

    phase = traj.add_phase('phase0', dm.Phase(ode_class=BreakwellODE, transcription=t))

    phase.set_time_options(fix_initial=True, fix_duration=True, units='s')

    phase.add_state('v', fix_initial=True, rate_source='u', units='m/s')
    phase.add_state('x', fix_initial=True, rate_source='v', units='m', shape=(1, ))
    phase.add_state('J', fix_initial=True, rate_source='J_dot')

    phase.add_control('u', units='m/s**2', scaler=0.1, continuity=False, rate_continuity=False,
                      rate2_continuity=False, shape=(1, ))

    # Maximize distance travelled in one second.
    phase.add_objective('J', loc='final')

    phase.add_path_constraint('x', upper=0.1)
    phase.add_boundary_constraint('x', loc='final', equals=0.0)
    phase.add_boundary_constraint('v', loc='final', equals=-1.0)

    p.model.linear_solver = om.DirectSolver()

    p.setup(check=True, force_alloc_complex=True)

    phase.set_time_val(initial=0.0, duration=1.0)
    phase.set_state_val('x', [0, 0])
    phase.set_state_val('v', [1, 0])
    phase.set_state_val('J', [0, 4])
    phase.set_control_val('u', [-5, -5])

    return p


@use_tempdirs
class TestBreakwellExample(unittest.TestCase):

    @classmethod
    def tearDownClass(cls):
        for filename in ['total_coloring.pkl', 'SLSQP.out', 'SNOPT_print.out']:
            if os.path.exists(filename):
                os.remove(filename)

    @staticmethod
    def _assert_results(p, tol=1.0E-4):
        x = p.get_val('traj.phase0.timeseries.x')
        v = p.get_val('traj.phase0.timeseries.v')
        J = p.get_val('traj.phase0.timeseries.J')
        J_expected = 2*v[0]**3/(9*(0.1 - x[0])) - 2*v[-1]**3/(9*(0.1-x[-1]))

        assert_near_equal(x[-1], 0.0, tolerance=tol)
        assert_near_equal(v[-1], -1.0, tolerance=tol)
        assert_near_equal(J[-1], J_expected, tolerance=tol)

    def test_ex_double_integrator_gl_compressed(self):
        p = double_integrator_direct_collocation('gauss-lobatto',
                                                 compressed=True)
        dm.run_problem(p)
        self._assert_results(p)

    def test_ex_double_integrator_gl_uncompressed(self):
        p = double_integrator_direct_collocation('gauss-lobatto',
                                                 compressed=False)
        dm.run_problem(p)
        self._assert_results(p)

    def test_ex_double_integrator_radau_compressed(self):
        p = double_integrator_direct_collocation('radau-ps',
                                                 compressed=True, optimizer='IPOPT')
        dm.run_problem(p)
        self._assert_results(p)

    def test_ex_double_integrator_radau_uncompressed(self):
        p = double_integrator_direct_collocation('radau-ps',
                                                 compressed=False)
        dm.run_problem(p)
        self._assert_results(p)

    def test_ex_double_integrator_birkhoff(self):
        p = double_integrator_direct_collocation('birkhoff', optimizer='IPOPT')
        dm.run_problem(p)
        self._assert_results(p)

    def test_check_partials(self):
        p = double_integrator_direct_collocation('birkhoff', optimizer='SLSQP')
        p.run_model()
        cpd = p.check_partials(method='cs', out_stream=None)
        assert_check_partials(cpd)


if __name__ == "__main__":
    unittest.main()



from collections import ChainMap
import datetime
import itertools
from pathlib import Path
import os.path

import numpy as np

from dymos.trajectory.trajectory import Trajectory
from dymos.phase.phase import Phase

try:
    from bokeh.io import save
    from bokeh.layouts import column, grid, row
    from bokeh.models import Legend, DataRange1d, DataTable, Div, \
        ColumnDataSource, TableColumn, TabPanel, Tabs, \
        CheckboxButtonGroup, CustomJS, MultiChoice
    from bokeh.plotting import figure, curdoc
    import bokeh.palettes as bp
    import bokeh.resources as bokeh_resources
    _NO_BOKEH = False
except ImportError:
    _NO_BOKEH = True

import openmdao.api as om
from openmdao.utils.units import conversion_to_base_units
from openmdao.utils.mpi import MPI
from openmdao.recorders.sqlite_recorder import META_KEY_SEP


# Javascript Callback when the solution/simulation checkbox buttons are toggled
# args: (figures)
_js_show_figures = """
const phases_to_show = phase_select.value;
const kinds_to_show = sol_sim_toggle.active;
var figures = figures;
var renderer;
var renderer_phase;

function show_renderer(renderer, phases_to_show, kinds_to_show) {
    var tags = renderer.tags;
    for(var k=0; k < tags.length; k++) {
        if (tags[k].substring(0, 6) == 'phase:') {
            renderer_phase = tags[k].substring(6);
            break;
        }
    }
    return ((tags.includes('sol') && kinds_to_show.includes(0)) ||
            (tags.includes('sim') && kinds_to_show.includes(1))) &&
           phases_to_show.includes(renderer_phase);
}

for (var i = 0; i < figures.length; i++) {
    if (figures[i]) {
        for (var j=0; j < figures[i].renderers.length; j++) {
            renderer = figures[i].renderers[j];
            // Get the phase with which this renderer is associated
            renderer.visible = show_renderer(renderer, phases_to_show, kinds_to_show);
        }
    }
}
"""


def _meta_tree_subsys_iter(tree, recurse=True, cls=None, path=None):
    """
    Yield a generator of local subsystems of this system.

    Parameters
    ----------
    recurse : bool
        If True, iterate over the whole tree under this system.
    cls : None, str, or Sequence
        The class of the nodes to be iterated
    path : The absolute path of the given tree.

    Yields
    ------
    type or None
    """
    _cls = [cls] if isinstance(cls, str) else cls

    for s in tree['children']:
        if s['type'] != 'subsystem':
            continue
        else:
            s['path'] = f'{path}.{s["name"]}' if path else s['name']

        if cls is None or s['class'] in _cls:
            yield s
        if recurse:
            for child in _meta_tree_subsys_iter(s, recurse=True, cls=_cls, path=s['path']):
                yield child


def _get_model_options_from_cr(cr, syspath, run_number=None):
    """Retrieve model options for the given system from the given case reader.

    The the options for system stored in the given case reader.
    If there is more than one set of model options, this function returns the last recorded ones.

    Parameters
    ----------
    cr : CaseReader
        The CaseReader instance holding the data.
    syspath : str
        Pathname of system whose options are requested.
    run_number : int or None
        Run_driver or run_model iteration to inspect, if given. If None, return the last available.

    Returns
    -------
    dict{system: {key: val}}
        The model options dictionary for the given system.
    """
    if not cr._system_options:
        raise RuntimeError('System options were not recorded.')

    # need to handle edge case for v11 recording
    if cr._format_version < 12:
        SEP = '_'
    else:
        SEP = META_KEY_SEP

    for key in cr._system_options:
        if key.find(SEP) > 0:
            name, num = key.rsplit(SEP, 1)
        else:
            name = key
            num = 0

        # Get the component associated with the highest run number
        if name == syspath and (run_number is None or run_number == int(num)):
            return cr._system_options[key]['component_options']
    else:
        raise KeyError(f'No options found for system {syspath}')


def _get_traj_and_phases_from_problem(problem, rank: int = 0):
    """Retrieve a dictionary tree structure of all trajectories and phases in the problem.

    Parameters
    ----------
    problem : Problem
        The problem being searched for trajectories and phases.
    rank : int
        Under MPI, the returned trajectory tree is only valid on this rank.

    Returns
    -------
    trajs : dict
        A dictionary for each trajectory containing its parameter options dictionary and
        a subdictionary of phases and their associated options. Under MPI, this dictionary
    """
    comm_rank = 0 if MPI is None else MPI.COMM_WORLD.rank
    traj_options = _gather_system_options(problem.model, ['parameter_options'],
                                          sys_cls=Trajectory, rank=rank)
    phase_options = _gather_system_options(problem.model,
                                           ['time_options', 'parameter_options',
                                            'state_options', 'control_options'],
                                           sys_cls=Phase, rank=rank)

    trajs = {}

    if comm_rank == 0:
        for traj_path, toptions in traj_options.items():
            trajs[traj_path] = {'phases': {},
                                'parameter_options': toptions['parameter_options'],
                                'name': traj_path}
            for phase_path, poptions in phase_options.items():
                if phase_path.startswith(f'{traj_path}.'):
                    trajs[traj_path]['phases'][phase_path] = {'name': phase_path.split('.')[-1]}
                    for opt in ('time_options', 'parameter_options', 'state_options',
                                'control_options'):
                        trajs[traj_path]['phases'][phase_path][opt] = poptions[opt]

    return trajs


# TODO: Enable this function when system options are correctly stored under MPI
def _get_trajs_and_phases_from_cr(cr, problem=None):  # pragma: no cover
    """Retrieve dictionaries of the trajectories and phases from the given case reader and problem.

    Due to a bug in OpenMDAO, model options may not be available from the case reader. In this case,
    use the problem to obtain them.

    Parameters
    ----------
    cr : CaseReader
        The CaseReader from which the model tree should be loaded.
    problem : Problem, optional
        The Problem instance from which system options should be loaded as a fallback.

    Returns
    -------
    _type_
        _description_
    """
    trajs = {}
    traj_cls = 'dymos.trajectory.trajectory:Trajectory'
    phase_cls = ['dymos.phase.phase:Phase',
                 'dymos.phase.phase:AnalyticPhase',
                 'dymos.phase.phase:SimualationPhase']

    traj_nodes = [n for n in _meta_tree_subsys_iter(cr.problem_metadata['tree'], cls=traj_cls)]
    phase_nodes = [n for n in _meta_tree_subsys_iter(cr.problem_metadata['tree'], cls=phase_cls)]

    for tn in traj_nodes:
        phase_nodes = [n for n in _meta_tree_subsys_iter(tn, cls=phase_cls, path=tn['path'])]
        traj_path = tn['path']
        trajs[traj_path] = {'phases': {}, 'parameter_options': {}, 'name': tn['name']}

        traj_options = _get_model_options_from_cr(cr=cr, syspath=traj_path)

        if MPI and MPI.COMM_WORLD.rank == 0:
            for param_name, param_options in traj_options['parameter_options'].items():
                trajs[traj_path]['parameter_options'][param_name] = param_options
            for pn in phase_nodes:
                phase_path = pn['path']
                phase_options = _get_model_options_from_cr(cr=cr, syspath=phase_path)
                phase_meta = trajs[traj_path]['phases'][phase_path] = {'time_options': None,
                                                                       'parameter_options': None,
                                                                       'state_options': None,
                                                                       'control_options': None,
                                                                       'name': pn['name']}
                phase_meta['time_options'] = phase_options['time_options']
                phase_meta['parameter_options'] = phase_options['parameter_options']
                phase_meta['state_options'] = phase_options['state_options']
                phase_meta['control_options'] = phase_options['control_options']

    return trajs


def _load_data_sources(traj_and_phase_meta=None, solution_record_file=None, simulation_record_file=None):
    """
    Load the data for the timeseries plots from the given solution and record files.

    Parameters
    ----------
    traj_and_phase_meta : dict
        A tree dictionary structure that contains the trajectories, their child phases,
        and associated options.
    solution_record_file : str
        The path to the solution record file.
    sim_record_file : str
        The path to the corresponding simulation record file.

    Returns
    -------
    dict
        A dictionary containing parameter data, solution timeseries data, simulation timeseries data, and
        units for each timeseries output.
    """
    data_dict = {}

    if solution_record_file is not None and Path(solution_record_file).is_file():
        sol_cr = om.CaseReader(solution_record_file)
        sol_case = sol_cr.get_case('final')
    else:
        sol_case = None

    if simulation_record_file is not None and Path(simulation_record_file).is_file():
        sim_cr = om.CaseReader(simulation_record_file)
        sim_case = sim_cr.get_case('final')
    else:
        sim_case = None

    source_case = sol_case or sim_case or None
    outputs = {abs_path: meta for abs_path, meta in source_case.list_outputs(out_stream=None, units=True)}

    if source_case is None:
        om.issue_warning('No recorded data provided. Trajectory results report will not be created.')
        return

    if sol_case:
        sol_outputs = {abs_path: meta for abs_path, meta in
                       sol_case.list_outputs(out_stream=None, units=True) if 'timeseries' in abs_path}
    else:
        sol_outputs = None

    if sim_case:
        sim_outputs = {abs_path: meta for abs_path, meta in
                       sim_case.list_outputs(out_stream=None, units=True) if 'timeseries' in abs_path}
    else:
        sim_outputs = None

    source_case = sol_case or sim_case
    outputs = sol_outputs or sim_outputs

    if sim_outputs is not None and sol_outputs is not None:
        if not set(sim_outputs.keys()).issubset(sol_outputs.keys()):
            om.issue_warning('Simulation file does not contain the same outputs as the solution '
                             'file. Skipping plotting of simulation timeseries data.\nThe following '
                             'outputs are in the simulation results but not in the solution results:\n'
                             f'{set(sim_outputs.keys()) - set(sol_outputs.keys())}')
            sim_case = None

    for traj_path, traj_data in traj_and_phase_meta.items():
        data_dict[traj_path] = {'param_data_by_phase': {},
                                'sol_data_by_phase': {},
                                'sim_data_by_phase': {},
                                'timeseries_units': {}}

        for phase_path, phase_data in traj_data['phases'].items():
            phase_name = phase_data['name']
            phase_param_data = \
                data_dict[traj_path]['param_data_by_phase'][phase_name] = \
                {'param': [], 'val': [], 'units': []}

            phase_sol_data = data_dict[traj_path]['sol_data_by_phase'][phase_name] = {}
            phase_sim_data = data_dict[traj_path]['sim_data_by_phase'][phase_name] = {}

            ts_outputs = {op: meta for op, meta in outputs.items()
                          if op.startswith(f'{phase_path}.timeseries.')}

            # Populate the phase parameter data
            phase_params = traj_and_phase_meta[traj_path]['phases'][phase_path]['parameter_options']
            for param_name in sorted(phase_params, key=str.casefold):
                units = phase_params[param_name]['units']
                param_path = f'{traj_path}.{phase_name}.parameter_vals:{param_name}'
                phase_param_data['param'].append(param_name)
                phase_param_data['units'].append(units)
                phase_param_data['val'].append(source_case.get_val(param_path, units=units))

            # Find the "largest" unit used for any timeseries output across all phases
            ts_units_dict = data_dict[traj_path]['timeseries_units']
            for abs_name in sorted(ts_outputs.keys(), key=str.casefold):
                meta = ts_outputs[abs_name]
                prom_name = meta['prom_name']
                var_name = prom_name.split('.')[-1]

                if var_name not in ts_units_dict:
                    ts_units_dict[var_name] = meta['units']
                else:
                    _, new_conv_factor = conversion_to_base_units(meta['units'])
                    _, old_conv_factor = conversion_to_base_units(ts_units_dict[var_name])
                    if new_conv_factor < old_conv_factor:
                        ts_units_dict[var_name] = meta['units']

            # Populate the phase timeseries data
            for output_name in sorted(ts_outputs.keys(), key=str.casefold):
                meta = ts_outputs[output_name]
                prom_name = meta['prom_name']
                var_name = prom_name.split('.')[-1]

                if sol_case:
                    data = sol_case.get_val(prom_name, units=ts_units_dict[var_name])
                    phase_sol_data[var_name] = data
                if sim_case:
                    data = sim_case.get_val(prom_name, units=ts_units_dict[var_name])
                    phase_sim_data[var_name] = data

    return data_dict


def _gather_system_options(model, options, sys_cls=None, rank=0,):
    """Retreive system options for systems of the given class and/or pathname.

    Parameters
    ----------
    model : System
        The root system from which model options are being gathered.
    options : Sequence of str
        The names of the options to be gathered from systems.
        Should be one of Trajectory or Phase.
    sys_cls : class, optional
        The class of system for which we want to retrieve options, or None if we
        should look for all systems regardless of class.
    rank : int
        The rank onto which the system options shoud be gathered. The retured
        dictionary of system options will only be valid on this rank.
    """
    comm_size = 1 if MPI is None else MPI.COMM_WORLD.size
    comm_rank = 0 if MPI is None else MPI.COMM_WORLD.rank

    system_options = {}
    for subsys in model.system_iter(include_self=True, recurse=True, typ=sys_cls):
        system_options[subsys.pathname] = {k: v['val'] for k, v in subsys.options._dict.items()
                                           if k in options}

    if comm_size > 1:
        gathered = MPI.COMM_WORLD.gather(system_options, rank)

        if comm_rank == 0:
            system_options.update(dict(ChainMap(*gathered)))

    return system_options


def _new_figure(x_name, y_name, x_units, y_units, margin, x_range=None):
    fig_kwargs = {'x_range': x_range} if x_range is not None else {}

    tool_tips = [(f'{x_name}', f'@{x_name}'), (f'{y_name}', f'@{y_name}')]

    fig = figure(tools='pan,box_zoom,xwheel_zoom,hover,undo,reset,save',
                 tooltips=tool_tips,
                 x_axis_label=f'{x_name} ({x_units})',
                 y_axis_label=f'{y_name} ({y_units})',
                 toolbar_location='above',
                 sizing_mode='stretch_both',
                 min_height=250, max_height=300,
                 margin=margin,
                 **fig_kwargs)
    fig.xaxis.axis_label_text_font_size = '10pt'
    fig.yaxis.axis_label_text_font_size = '10pt'
    fig.toolbar.autohide = True
    return fig


def make_timeseries_report(prob, solution_record_file=None, simulation_record_file=None,
                           x_name='time', ncols=2, margin=10, theme='light_minimal'):
    """
    Create the bokeh-based timeseries results report.

    Parameters
    ----------
    prob : om.Problem
        The problem instance for which the timeseries plots are being created.
    solution_record_file : str
        The path to the solution record file, if available.
    simulation_record_file : str
        The path to the simulation record file, if available.
    x_name : str
        Name of the horizontal axis variable in the timeseries.
    ncols : int
        The number of columns of timeseries output plots.
    margin : int
        A margin to be placed between the plot figures.
    theme : str
        A valid bokeh theme name to style the report.
    """
    comm_rank = 0 if MPI is None else MPI.COMM_WORLD.rank
    report_dir = Path(prob.get_reports_dir()) if prob is not None else Path(os.getcwd())
    if not report_dir.exists():
        report_dir = Path(os.getcwd())

    if prob is not None:
        # Retrieve system information fom the problem if available.
        traj_data = _get_traj_and_phases_from_problem(prob)

    # The rest of the traj report generation occurs on rank 0
    if comm_rank == 0:

        # For the primary timeseries in each phase in each trajectory, build a set of the pathnames
        # to be plotted.
        source_data = _load_data_sources(traj_data, solution_record_file, simulation_record_file)

        # Colors of each phase in the plot. Start with the bright colors followed by the faded ones.
        if not _NO_BOKEH:
            colors = bp.d3['Category20'][20][0::2] + bp.d3['Category20'][20][1::2]
            curdoc().theme = theme

        for traj_path, traj_data in source_data.items():
            report_filename = f'{traj_path}_results_report.html'
            report_path = report_dir / report_filename
            if _NO_BOKEH:
                with open(report_path, 'wb') as f:
                    f.write("<html>\n<head>\n<title> \nError: bokeh not available</title>\n</head> <body>\n"
                            "This report requires bokeh but bokeh was not available in this python installation.\n"
                            "</body></html>".encode())
                continue

            param_tables = []
            phase_names = [k.split('.')[-1] for k in traj_data['param_data_by_phase']]
            for phase_name, param_data in traj_data['param_data_by_phase'].items():
                # Make the parameter table
                columns = [
                    TableColumn(field='param', title='Parameter'),
                    TableColumn(field='val', title='Value'),
                    TableColumn(field='units', title='Units'),
                ]
                param_tables.append(DataTable(source=ColumnDataSource(param_data),
                                              columns=columns,
                                              index_position=None,
                                              height=30*len(param_data['param']),
                                              sizing_mode='stretch_both'))

            # Plot the timeseries
            ts_units_dict = source_data[traj_path]['timeseries_units']

            figures = {}
            legend_data_per_figure = {}
            x_range = None

            # var_name is the actual dymos variable name, without any index information.
            # var_name_with_idxs is the variable name with index information.

            for var_name in sorted(ts_units_dict.keys(), key=str.casefold):
                for i, phase_name in enumerate(phase_names):
                    color = colors[i % 20]
                    sol_data = source_data[traj_path]['sol_data_by_phase'][phase_name]
                    sim_data = source_data[traj_path]['sim_data_by_phase'][phase_name]

                    if x_name in sol_data and var_name in sol_data:
                        shape = sol_data[var_name].shape[1:]
                        indices = list(itertools.product(*(range(dim) for dim in shape)))
                        if np.prod(shape) > 1:
                            sources = {}
                            for idxs in indices:
                                str_idxs = ','.join([str(i) for i in idxs])
                                # Bokeh ColumnDataSource doesn't allow special characters in keys,
                                # but we want the y_axis label to show the indices of the columns
                                # being plotted as 'varname[i,j,k]'.
                                sources[f'{var_name}[{str_idxs}]'] = s = f'{var_name}_{str_idxs.replace(",", "_")}'
                                sol_data_column = sol_data[var_name][(slice(None), *idxs)]
                                sol_data[s] = sol_data_column
                                if sim_data:
                                    sim_data_column = sim_data[var_name][(slice(None), *idxs)]
                                    sim_data[s] = sim_data_column
                        else:
                            sources = {var_name: var_name}
                        sol_source = ColumnDataSource(sol_data)
                        if sim_data:
                            sim_source = ColumnDataSource(sim_data)

                        for var_name_with_idxs, _source in sources.items():
                            legend_items = []

                            if var_name_with_idxs in figures:
                                fig = figures[var_name_with_idxs]
                                fig_legend_data = legend_data_per_figure[var_name_with_idxs]
                            else:
                                fig = _new_figure(x_name=x_name, y_name=var_name_with_idxs,
                                                  x_units=ts_units_dict[x_name],
                                                  y_units=ts_units_dict[var_name],
                                                  margin=margin,
                                                  x_range=x_range)
                                figures[var_name_with_idxs] = fig
                                fig_legend_data = legend_data_per_figure[var_name_with_idxs] = []
                            if sol_data:
                                sol_plot = fig.scatter(x=x_name, y=_source, source=sol_source,
                                                       color=color, size=5)
                                sol_plot.tags.extend(['sol', f'phase:{phase_name}'])
                                legend_items.append(sol_plot)
                            if sim_data:
                                sim_plot = fig.line(x=x_name, y=_source, source=sim_source, color=color)
                                sim_plot.tags.extend(['sim', f'phase:{phase_name}'])
                                legend_items.append(sim_plot)
                            fig_legend_data.append((phase_name, legend_items))
                            if x_range is None:
                                x_range = fig.x_range
                            fig.y_range = DataRange1d(min_interval=1.0E-12)

                            figures[var_name_with_idxs] = fig

            # Add the legend data to each figure
            for var_name_with_idxs, fig in figures.items():
                legend_data = legend_data_per_figure[var_name_with_idxs]
                legend = Legend(items=legend_data, location='center', label_text_font_size='8pt',
                                orientation='vertical')
                fig.add_layout(legend, 'right')

            # Since we're putting figures in two columns, make sure we have an even number of things to put in the layout.
            if len(figures) % 2 == 1:
                figures['__NONE__'] = None

            param_panels = [TabPanel(child=table, title=f'{phase_names[i]} parameters')
                            for i, table in enumerate(param_tables)]

            sol_sim_toggle = CheckboxButtonGroup(labels=['Solution', 'Simulation'], active=[0, 1])

            sol_sim_row = row(children=[Div(text='Display data:', sizing_mode='stretch_height'),
                                        sol_sim_toggle],
                              sizing_mode='stretch_both',
                              max_height=50)

            phase_select = MultiChoice(options=[phase_name for phase_name in phase_names],
                                       value=[phase_name for phase_name in phase_names],
                                       sizing_mode='stretch_both',
                                       min_width=400, min_height=50)

            phase_select_row = row(children=[Div(text='Plot phases:'), phase_select],
                                   sizing_mode='stretch_width')

            figures_grid = grid(children=list(figures.values()), ncols=ncols, sizing_mode='stretch_both')

            ts_layout = column(children=[sol_sim_row, phase_select_row, figures_grid],
                               sizing_mode='stretch_both')

            tab_panes = Tabs(tabs=[TabPanel(child=ts_layout, title='Timeseries')] + param_panels,
                             sizing_mode='stretch_both',
                             active=0)

            summary = rf'Results of {prob._name}<br>Creation Date: {datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")}'

            report_layout = column(children=[Div(text=summary), tab_panes], sizing_mode='stretch_both')

            # Assign callbacks

            sol_sim_toggle.js_on_change("active",
                                        CustomJS(code=_js_show_figures,
                                                 args=dict(figures=figures,
                                                           sol_sim_toggle=sol_sim_toggle,
                                                           phase_select=phase_select)))

            phase_select.js_on_change("value",
                                      CustomJS(code=_js_show_figures,
                                               args=dict(figures=figures,
                                                         sol_sim_toggle=sol_sim_toggle,
                                                         phase_select=phase_select)))

            # Save
            save(report_layout, filename=report_path, title=f'trajectory results for {traj_path}',
                 resources=bokeh_resources.INLINE)



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

import dymos as dm
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE
from dymos.grid_refinement.error_estimation import eval_ode_on_grid, compute_state_quadratures

from openmdao.utils.general_utils import set_pyoptsparse_opt
OPT, OPTIMIZER = set_pyoptsparse_opt('SLSQP', fallback=True)


@use_tempdirs
class TestBrachistochroneExample(unittest.TestCase):

    @require_pyoptsparse(optimizer='SLSQP')
    def _run_brachistochrone(self, transcription_class=dm.Radau, control_type='control', g=9.80665):
        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = OPTIMIZER
        p.driver.declare_coloring(tol=1.0E-12)

        tx = transcription_class(num_segments=15,
                                 order=5,
                                 compressed=False)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)
        phase.timeseries_options['use_prefix'] = True
        phase.timeseries_options['include_state_rates'] = True
        p.model.add_subsystem('traj0', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)
        phase.add_state('y', fix_initial=True, fix_final=True)

        # Note that by omitting the targets here Dymos will automatically attempt to connect
        # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
        phase.add_state('v', fix_initial=True, fix_final=False)

        if control_type == 'control':
            phase.add_control('theta', continuity=True, rate_continuity=True,
                              units='deg', lower=0.01, upper=179.9)
        elif control_type == 'polynomial_control':
            phase.add_control('theta', units='deg', lower=0.01, upper=179.9, order=3, control_type='polynomial')

        phase.add_parameter('g', units='m/s**2', val=1.0)

        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        phase.timeseries_options['include_state_rates'] = True

        p.setup()

        phase.set_time_val(0, 2)

        phase.set_state_val('x', [0, 10], units='m')
        phase.set_state_val('y', [10, 5], units='m')
        phase.set_state_val('v', [0, 5], units='m/s')
        phase.set_control_val('theta', [5, 90], units='deg')
        phase.set_parameter_val('g', g)

        return p

    def test_compute_state_quadratures(self):
        """ Tests that the eval_ode_on_grid method works correctly against a variety of problems. """

        for tx_class in (dm.Radau, dm.GaussLobatto):
            if tx_class is dm.Radau:
                err_tol = 1.0E-4
            else:
                err_tol = 1.0E-3
            for control_type in ('control', 'polynomial_control'):
                for g in (9.80665, 1.62):
                    with self.subTest(msg=f'{tx_class.__name__} - {control_type} - g = {g}'):
                        p = self._run_brachistochrone(transcription_class=tx_class,
                                                      control_type=control_type,
                                                      g=g)
                        p.run_driver()

                        phase = p.model.traj0.phases.phase0

                        x, u, pp, f = eval_ode_on_grid(phase, phase.options['transcription'])

                        t_duration = phase.get_val('t_duration')

                        x_hat = compute_state_quadratures(x, f, t_duration, phase.options['transcription'])

                        # Check that the computed state rates in the grid refinement ODE are equal to those
                        # in the original problem. (on the same grid they should be within machine precision)

                        print(f'{tx_class.__name__} - {control_type} - g = {g}')

                        for name, options in phase.control_options.items():
                            u_solution = phase.get_val(f'timeseries.controls:{name}')
                            print(f'{name} interpolation error',
                                  max(np.abs(u[name].ravel() - u_solution.ravel())))

                        for name, options in phase.state_options.items():
                            x_solution = phase.get_val(f'timeseries.states:{name}')
                            f_solution = phase.get_val(f'timeseries.state_rates:{name}')

                            print(f'{name} interpolation error', max(np.abs(x[name].ravel() - x_solution.ravel())))
                            print(f'{name} rate error', max(np.abs(f[name].ravel() - f_solution.ravel())))
                            print(f'{name} error:', max(np.abs(x_hat[name] - x[name])))

                            assert_near_equal(x[name].ravel(), x_solution.ravel())
                            assert_near_equal(f[name].ravel(), f_solution.ravel())
                            assert_near_equal(x_hat[name], x[name], tolerance=err_tol)
